"""Module for loading vehicles, scenarios, running them and managing them"""

import ast
from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple

import fastsim
import numpy as np
import pandas as pd
from fastsim import cycle, simdrive, vehicle
import os

# for debugging convenience
from typing_extensions import Self

from t3co.objectives import accel, fueleconomy, gradeability
from t3co.run import Global as gl
from t3co.tco import tco_analysis

# import importlib
# tco_analysis = importlib.reload(tco_analysis)


# --------------------------------- \\ powertrain adjustment methods --------------------------------- #
@dataclass
class Config:
    """
    New class to read T3COConfig file containing analysis attributes like vehicle and scenario paths, and scenario attribute overrides

    """
    analysis_id: int = 0
    analysis_name: str = ""
    vehicle_file: str = ""
    scenario_file: str = ""
    dst_dir: str = ""
    write_tsv: bool = False
    selections: str = ""
    # selections: list = field(default_factory=list)
    vehicle_life_yr: float = 0

    # Fueling
    ess_max_charging_power_kw: float = 0
    fs_fueling_rate_kg_per_min: float = 0
    fs_fueling_rate_gasoline_gpm: float = 0
    fs_fueling_rate_diesel_gpm: float = 0

    # Optimization
    algorithms: str = ""
    lw_imp_curves: str = ""
    eng_eff_imp_curves: str = ""
    aero_drag_imp_curves: str = ""
    lw_imp_curve_sel: str = ""
    eng_eff_imp_curve_sel: str = ""
    aero_drag_imp_curve_sel: str = ""
    skip_all_opt: bool = True
    constraint_range: bool = False
    constraint_accel: bool = False
    constraint_grade: bool = False
    objective_tco: bool = False
    constraint_c_rate: bool = False
    constraint_trace_miss_dist_percent_on: bool = False
    objective_phev_minimize_fuel_use: bool = False

    # Opportunity Cost
    activate_tco_payload_cap_cost_multiplier: bool = False
    activate_tco_fueling_dwell_time_cost: bool = False
    fdt_frac_full_charge_bounds: list = field(default_factory=list)
    activate_mr_downtime_cost: bool = False

    def from_file(self, filename: str, analysis_id: int) -> Self:
        """
        This method generates a Config dictionary from CSV file and calls Config.from_dict

        Args:
            filename (str): path of input T3CO Config file
            analysis_id (int): analysis ID selections

        Returns:
            Self.from_dict: method that gets Config instance from config_dict
        """
        filename = str(filename)

        config_df = pd.read_csv(filename, index_col="analysis_id").loc[analysis_id]
        config_dict = config_df.to_dict()

        return self.from_dict(config_dict=config_dict)

    def from_dict(self, config_dict: dict) -> Self:
        """
        This method generates a Config instance from config_dict

        Args:
            config_dict (dict): dictionary containing fields from T3CO Config input CSV file

        Returns:
            Self: Config instance containining all values from T3CO Config CSV file
        """
        # config_dict['selections'] = np.array(ast.literal_eval(config_dict['selections']))
        try:
            config_dict["selections"] = ast.literal_eval(config_dict["selections"])
        except:  # noqa: E722
            config_dict["selections"] = int(config_dict["selections"])
        self.__dict__.update(config_dict)

    def validate_analysis_id(self, filename: str, analysis_id: int = 0) -> Self:
        """
        This method validates that correct analysis id is input

        Args:
            filename (str): T3CO Config input CSV file path

        Raises:
            Exception: Error if analysis_id not found
        """
        filename = str(filename)
        config_df = pd.read_csv(filename)
        print(f"Try these analysis IDs instead: {list(config_df['analysis_id'])}")
        assert (
            analysis_id in config_df["analysis_id"]
        ), "Given analysis_id not in config input file"
        raise Exception


@dataclass
class Scenario:
    """
    Class object that contains all TCO parameters and performance target (range, grade, accel) information \
        for a vehicle such that performance and TCO can be computed during optimization
    """

    selection: float = 0
    drive_cycle: str = ""
    use_config: bool = True
    vmt_reduct_per_yr: float = 0
    vmt: list = field(default_factory=list)
    constant_trip_distance_mi: float = 0
    vehicle_life_yr: float = 0
    desired_ess_replacements: float = 0
    discount_rate_pct_per_yr: float = 0

    ess_max_charging_power_kw: float = 0
    ess_cost_dol_per_kw: float = 0
    ess_cost_dol_per_kwh: float = 0
    ess_base_cost_dol: float = 0
    ess_cost_reduction_dol_per_yr: float = 0
    ess_salvage_value_dol: float = 0
    ess_charge_rate_kW: float = 0
    pe_mc_cost_dol_per_kw: float = 0
    pe_mc_base_cost_dol: float = 0
    fc_ice_cost_dol_per_kw: float = 0
    fc_ice_base_cost_dol: float = 0
    fc_fuelcell_cost_dol_per_kw: float = 0
    fs_cost_dol_per_kwh: float = 0
    fs_h2_cost_dol_per_kwh: float = 0
    plug_base_cost_dol: float = 0
    markup_pct: float = 0
    tax_rate_pct: float = 0
    fc_cng_ice_cost_dol_per_kw: float = 0
    fs_cng_cost_dol_per_kwh: float = 0
    vehicle_glider_cost_dol: float = 0
    segment_name: str = ""
    gvwr_kg: float = 0
    gvwr_credit_kg: float = 0
    # a list of fuels, basecase fuel is singleton list
    fuel_type: list = field(default_factory=list)
    maint_oper_cost_dol_per_mi: list = field(default_factory=list)
    vocation: str = ""
    vehicle_class: str = ""
    model_year: float = 0
    region: str = ""
    target_range_mi: float = 0
    min_speed_at_6pct_grade_in_5min_mph: float = 0
    min_speed_at_1p25pct_grade_in_5min_mph: float = 0
    max_time_0_to_60mph_at_gvwr_s: float = 0
    max_time_0_to_30mph_at_gvwr_s: float = 0
    # TDA vars
    lw_imp_curve_sel: str = ""
    eng_eff_imp_curve_sel: str = ""
    aero_drag_imp_curve_sel: str = ""
    # computed vars
    # scenario_gge_regional_temporal_fuel_price: str = ""
    originalcargo_kg: float = (
        -1.0
    )  # if needed, should be assigned immediately after vehicle read in
    # For adding mass from CdA during optimization. veh_kg = glider_kg + powertrainKg, where
    # glider_kg is assigned the value of originalglider_kg + CdAKg
    originalglider_kg: float = -1.0
    # for adding incremental cost to glider from different CdA guesses in moo loop
    originalGliderPrice: float = -1.0
    # for adding percent improvemnt cost to engine efficiency when optimizing CONV
    originalIceDolPerKw: float = -1.0
    # for adjusting fuel converter efficiency based on new peak eff
    origfc_eff_map: list = field(default_factory=list)
    # for adjusting drag coefficient of vehicle
    originaldrag_coef: float = -1

    ess_init_soc_grade: float = -1.0
    ess_init_soc_accel: float = -1.0

    soc_norm_init_for_accel_pct: float = -1
    soc_norm_init_for_grade_pct: float = -1

    # fuel storage
    fs_fueling_rate_gasoline_gpm: float = 0
    fs_fueling_rate_diesel_gpm: float = 0
    fs_fueling_rate_kg_per_min: float = 0

    ### PHEV stuff
    # UF for % of miles in charge depleting mode
    phev_utility_factor_override: float = -1
    phev_utility_factor_computed: float = -1
    # percent (fractional) of motor power for setting kw_fc_demand_on during optimization
    motor_power_override_kw_fc_demand_on_pct: float = -1

    # This will be used to figure out the number of miles travelled before needing to charge
    # must be greater than 0
    shifts_per_year: list = field(default_factory=list)

    missed_trace_correction: bool = False
    max_time_dilation: float = -1
    min_time_dilation: float = -1
    time_dilation_tol: float = -1

    #
    ### Optimization Settings
    #
    skip_opt: bool = False
    knob_min_ess_kwh: list = field(default_factory=list)
    knob_max_ess_kwh: list = field(default_factory=list)
    knob_min_motor_kw: list = field(default_factory=list)
    knob_max_motor_kw: list = field(default_factory=list)
    knob_min_fc_kw: list = field(default_factory=list)
    knob_max_fc_kw: list = field(default_factory=list)
    knob_min_fs_kwh: list = field(default_factory=list)
    knob_max_fs_kwh: list = field(default_factory=list)
    # placeholder for if max_c_rate need to be entered as parameters for each scenario
    # c_rate_kwh_array: list = field(default_factory=list)
    # c_rate_array: list = field(default_factory=list)
    objective_phev_minimize_fuel_use: bool = False
    constraint_c_rate: bool = False
    constraint_range: bool = False
    constraint_accel: bool = False
    constraint_grade: bool = False
    objective_tco: bool = False
    constraint_trace_miss_dist_percent_on: bool = False
    trace_miss_dist_percent: float = 0
    constraint_phev_minimize_fuel_use_on: bool = False
    constraint_phev_minimize_fuel_use_percent: float = 0

    #
    ### TCO Element Activations and vars
    #
    labor_rate_dol_per_hr: float = 0
    downtime_oppy_cost_dol_per_hr: float = 0

    # payload loss factor vars, PLF
    activate_tco_payload_cap_cost_multiplier: bool = True
    plf_ref_veh_empty_mass_kg: float = 0
    plf_scenario_vehicle_empty_kg: float = 0
    plf_reference_vehicle_cargo_capacity_kg: float = 0
    plf_scenario_vehicle_cargo_capacity_kg: float = 0  # includes cargo credit kg
    estimated_lost_payload_kg: float = 0

    # Fueling Dwell time factors, FDT
    activate_tco_fueling_dwell_time_cost: bool = False
    dlf_min_charge_time_hr: float = 0
    fdt_dwpt_fraction_power_pct: float = 0
    fdt_avg_overhead_hr_per_dwell_hr: float = 0
    fdt_frac_full_charge_bounds: float = 0
    fdt_num_free_dwell_trips: float = 0
    fdt_available_freetime_hr: float = 0
    # Insurance factors
    insurance_rates_pct_per_yr: list = field(default_factory=list)

    # Residual Rate
    residual_rate_pct: float = 0

    # Maintenance and Repair Downtime factors MR
    activate_mr_downtime_cost: bool = False
    mr_planned_downtime_hr_per_yr: float = 0
    mr_unplanned_downtime_hr_per_mi: list = field(default_factory=list)
    mr_avg_tire_life_mi: float = 0
    mr_tire_replace_downtime_hr_per_event: float = 0

    def from_config(self, config: Config = None) -> Self:
        """
        This method overrides certain scenario fields if use_config is True and config object is not None

        Args:
            config (Config, optional): Config object. Defaults to None.

        """
        fields_override = [
            "vehicle_life_yr",
            "fs_fueling_rate_kg_per_min",
            "fs_fueling_rate_gasoline_gpm",
            "fs_fueling_rate_diesel_gpm",
            "lw_imp_curve_sel",
            "eng_eff_imp_curve_sel",
            "aero_drag_imp_curve_sel",
            "constraint_range",
            "constraint_accel",
            "constraint_grade",
            "objective_tco",
            "constraint_c_rate",
            "constraint_trace_miss_dist_percent_on",
            "objective_phev_minimize_fuel_use",
            "activate_tco_payload_cap_cost_multiplier",
            "activate_tco_fueling_dwell_time_cost",
            "fdt_frac_full_charge_bounds",
            "activate_mr_downtime_cost",
        ]
        self.fields_overriden = []
        if self.use_config == True and config != None:
            for field_select in fields_override:
                if (config.__dict__[field_select] is not None) and (
                    not self.__dict__[field_select]
                ):
                    setattr(self, field_select, config.__getattribute__(field_select))
                    # print(f'field: {field}, type: {type(self.__getattribute__(field))}, value: {self.__getattribute__(field)}')
                    self.fields_overriden.append(field_select)
            print(f"Scenario Fields overridden from config: {self.fields_overriden}")
        else:
            print(
                f"Config file not attached or scenario.use_config set to False: {config}"
            )

        return self


# PHEV utility methods
def check_phev_init_socs(a_vehicle: vehicle.Vehicle, scenario: Scenario) -> None:
    """
    This function checks that soc_norm_init_for_grade_pct and soc_norm_init_for_accel_pct are present only for PHEVs

    Args:
        a_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        scenario (Scenario): T3CO scenario object
    """
    # these override ess_init_soc_grade and ess_init_soc_accel
    # these should ONLY be used for PHEV, for now, until discussed for use for HEV, BEV
    # init_soc = min_soc + (soc_norm_init_for_accel_pct * (max_soc - min_soc))
    if (
        np.isnan(scenario.soc_norm_init_for_grade_pct) == False
        and scenario.soc_norm_init_for_grade_pct != -1
    ):
        assert (
            a_vehicle.veh_pt_type == gl.PHEV
        ), "soc_norm_init_for_grade_pct only available for PHEVs"
        assert (
            scenario.ess_init_soc_grade == -1
        ), f"INPUT ERROR, user specifed ess_init_soc_grade {scenario.ess_init_soc_grade}, & soc_norm_init_for_grade_pct {scenario.soc_norm_init_for_grade_pct} for PHEV; the question of which one to use is ambiguous"
    if (
        np.isnan(scenario.soc_norm_init_for_accel_pct) == False
        and scenario.soc_norm_init_for_accel_pct != -1
    ):
        assert (
            a_vehicle.veh_pt_type == gl.PHEV
        ), "soc_norm_init_for_accel_pct only available for PHEVs"
        assert (
            scenario.ess_init_soc_accel == -1
        ), f"INPUT ERROR, user specifed ess_init_soc_accel {scenario.ess_init_soc_accel}, & soc_norm_init_for_accel_pct {scenario.soc_norm_init_for_accel_pct} for PHEV; the question of which one to use is ambiguous"


def get_phev_util_factor(
    scenario: Scenario, v: fastsim.vehicle.Vehicle, mpgge: dict
) -> float:
    """
    This function gets the PHEV utility factor derived from the computed range of the
    vehicle and the operational day range computed from shifts per year and the first vmt year

    Args:
        scenario (Scenario): T3CO scenario object
        v (fastsim.vehicle.Vehicle): FASTSim vehicle object
        mpgge (dict): Miles per Gallon Gasoline Equivalent dictionary

    Returns:
        uf (float): PHEV computed utility factor
    """
    uf = scenario.phev_utility_factor_override
    assert type(scenario.phev_utility_factor_override) in [
        int,
        float,
    ], "should be -1 or some float"
    cd_range_mi = fueleconomy.get_range_mi(mpgge, v, scenario)["cd_aer_phev_range_mi"]

    if uf == -1:
        shift_range_mi = scenario.vmt[0] / scenario.shifts_per_year
        scenario.phev_utility_factor_computed = round(
            min(shift_range_mi, cd_range_mi) / shift_range_mi, 3
        )
        uf = scenario.phev_utility_factor_computed
    return uf


# utility methods sim drives
def get_objective_simdrive(
    analysis_vehicle: vehicle.Vehicle, cycle: fastsim.cycle.Cycle
) -> fastsim.simdrive.SimDrive:
    """
    This function obtains the SimDrive for accel and grade test

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        cycle (fastsim.cycle.Cycle): FASTSim Cycle object

    Returns:
        sd (fastsim.simdrive.SimDrive): FASTSim SimDrive object containing vehicle inputs and simulation output attributes
    """
    sd = simdrive.SimDrive(cycle, analysis_vehicle)
    sd = sd.to_rust()
    sim_params = sd.sim_params
    sim_params.reset_orphaned()
    sim_params.missed_trace_correction = False
    # accel and grade traces are not achievable for our vehicles in the way we've constructed the tests, so suppress this warning with large tolerance
    sim_params.trace_miss_speed_mps_tol = np.inf
    sim_params.energy_audit_error_tol = np.inf
    sim_params.trace_miss_dist_tol = np.inf
    sd.sim_params = sim_params

    return sd


def run_grade_or_accel(
    test: str,
    analysis_vehicle: fastsim.vehicle.Vehicle,
    sim_drive: fastsim.simdrive.SimDrive,
    ess_init_soc: float,
) -> None:
    """
    This function handles initial SOC considerations for grade and accel tests

    If ess_init_soc override is passed, use that
    Else if the vehicle is an HEV, use the standard HEV init SOC values for accel and grade
    Else, let FASTSim determine init SOC in sim_drive()
        BEVs use max_soc
        PHEVs use max_soc
        Conv init_soc doesn't matter
        HEVs attempt SOC balancing but that is overrident by HEV test init SOC

    Args:
        test (str): 'accel' or 'grade' test
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        sim_drive (fastsim.simdrive.SimDrive): FASTSim SimDrive object
        ess_init_soc (float): ESS initial state of charge (SOC)

    Raises:
        Exception: if test not in ['accel', 'grade']
    """

    if test == "accel":
        # this is what SimAccelTest object was doing in previous version (non-Rust JIT)
        hev_init_soc = (analysis_vehicle.max_soc + analysis_vehicle.min_soc) / 2.0
    elif test == "grade":
        hev_init_soc = analysis_vehicle.min_soc
    else:
        raise Exception("this should not have happened")

    if ess_init_soc is not None:
        sim_drive.sim_drive_walk(ess_init_soc)
    elif analysis_vehicle.veh_pt_type == gl.HEV:
        sim_drive.sim_drive_walk(hev_init_soc)
    else:
        sim_drive.sim_drive()


# utility methods to create fastsim vehicles
def create_fastsim_vehicle(veh_dict: dict = None) -> fastsim.vehicle.Vehicle:
    """
    This function creates and returns an empty FASTSim vehicle object with no attributes or

    Args:
        veh_dict (dict, optional): Vehicle attributes dict. Defaults to None.

    Returns:
        v (fastsim.vehicle.Vehicle): FASTSim vehicle object
    """

    if not veh_dict:
        veh_dict = {"DELETEME": 0}
    v = vehicle.Vehicle(veh_dict=veh_dict)
    try:
        if not veh_dict:
            del v.DELETEME
    except AttributeError:
        pass
    return v


def get_vehicle(veh_no: int, veh_input_path: str) -> fastsim.vehicle.Vehicle:
    """
    This function loads vehicle object from vehicle number and input csv filepath

    Args:
        veh_no (int): vehicle selection number
        veh_input_path (str): vehicle model assumptions input CSV file path

    Returns:
        veh (fastsim.vehicle.Vehicle): FASTSim vehicle object
    """
    veh = vehicle.Vehicle.from_vehdb(veh_no, veh_input_path, to_rust=True)
    veh.set_derived()
    veh.set_veh_mass()

    return veh


# \\ end \\ utility methods to create fastsim vehicles


def get_scenario_and_cycle(
    veh_no: int,
    scenario_inputs_path: str,
    a_vehicle: fastsim.vehicle.Vehicle = None,
    config: Config = None,
) -> Tuple[Scenario, fastsim.cycle.Cycle]:
    """
    This function uses helper methods load_scenario and load_design_cycle_from_scenario \
        to get scenario object and cycle object corresponding to selected vehicle (by veh_no)

    Args:
        veh_no (int): vehicle selection number
        scenario_inputs_path (str): input file path for scenario assumptions CSV
        a_vehicle (fastsim.vehicle.Vehicle, optional): FASTSim vehicle object for given selection. Defaults to None.
        config (Config, optional): Config object for current analysis. Defaults to None.

    Returns:
        scenario (Scenario): T3CO scenario object selected
        cyc (fastsim.cycle.Cycle): FASTSim cycle object selected
    """
    scenario = load_scenario(veh_no, scenario_inputs_path, a_vehicle, config)
    cyc = load_design_cycle_from_scenario(scenario, gl.OPTIMIZATION_DRIVE_CYCLES)

    return scenario, cyc


def load_scenario(
    veh_no: int,
    scenario_inputs_path: str,
    a_vehicle: fastsim.vehicle.Vehicle = None,
    config: Config = None,
) -> Scenario:
    """
    This function gets the Scenario object from scenario input CSV filepath, initializes some fields,\
          and overrides some fields based on Config object

    Args:
        veh_no (int): vehicle selection number
        scenario_inputs_path (str): input file path for scenario assumptions CSV
        a_vehicle (fastsim.vehicle.Vehicle, optional): FASTSim vehicle object for given selection. Defaults to None.
        config (Config, optional): Config object for current analysis. Defaults to None.

    Returns:
        scenario (Scenario): Scenario object for given selection
    """
    scenarios = pd.read_csv(scenario_inputs_path)
    assert (
        len(scenarios[scenarios["selection"] == veh_no]) == 1
    ), f"conflict in {__file__}get_scenario(_): Scenario numbers in {scenario_inputs_path} are not unique "
    scenario_dict = scenarios[scenarios["selection"] == veh_no].to_dict("list")
    scenario_dict = {k: v[0] for k, v in scenario_dict.items()}
    scenario_dict["vehicle_class"] = " "
    scenario_dict["vehicle_class"] = (
        scenario_dict["vehicle_class"]
        .join(scenario_dict["scenario_name"].split()[:3])
        .lower()
    )

    if "scenario_name" in scenario_dict:
        del scenario_dict["scenario_name"]

    # handle PHEV fuels list and UF list, convert to lists
    fuels = scenario_dict["fuel_type"]
    if "[" in fuels and "]" in fuels:
        fuels = ast.literal_eval(
            fuels
        )  # PHEV ["CD electricity", "CD diesel", "CS diesel"]
    else:
        fuels = [fuels]
    scenario_dict["fuel_type"] = fuels

    # handle vmt, turn into list
    scenario_dict["vmt"] = ast.literal_eval(scenario_dict["vmt"])
    scenario_dict["mr_unplanned_downtime_hr_per_mi"] = ast.literal_eval(
        scenario_dict["mr_unplanned_downtime_hr_per_mi"]
    )
    # if config: scenario_dict['config'] = config
    scenario = Scenario(**scenario_dict)
    scenario = scenario.from_config(config)

    # convert insurance rates string into float list
    scenario.insurance_rates_pct_per_yr = list(
        np.float_(scenario.insurance_rates_pct_per_yr.strip(" ][").split(","))
    )

    # validate some inputs, assign as -1 if not provided by user in input file
    if np.isnan(scenario.ess_init_soc_grade):
        scenario.ess_init_soc_grade = -1
    if np.isnan(scenario.ess_init_soc_accel):
        scenario.ess_init_soc_accel = -1
    if np.isnan(scenario.soc_norm_init_for_accel_pct):
        scenario.soc_norm_init_for_accel_pct = -1
    if np.isnan(scenario.soc_norm_init_for_grade_pct):
        scenario.soc_norm_init_for_grade_pct = -1

    # PHEV settings and checks
    if (
        np.isnan(scenario.phev_utility_factor_override)
        or scenario.phev_utility_factor_override is None
    ):
        scenario.phev_utility_factor_override = -1
        # we need non-None vmt and shifts_per_year since there is no phev_utility_factor_override provided
        assert (
            scenario.shifts_per_year not in [False, None, np.nan]
        ), f"invalid shifts_per_year value {scenario.shifts_per_year}, need a valid shifts_per_year (positive integer) value to compute utility factor since there is no phev_utility_factor_override provided"
        assert (
            scenario.vmt[0] is not None
        ), "we need non-None vmt since there is no phev_utility_factor_override provided"
    if (
        np.isnan(scenario.motor_power_override_kw_fc_demand_on_pct)
        or scenario.motor_power_override_kw_fc_demand_on_pct is None
    ):
        scenario.motor_power_override_kw_fc_demand_on_pct = -1
    elif scenario.motor_power_override_kw_fc_demand_on_pct != -1:
        assert (
            scenario.motor_power_override_kw_fc_demand_on_pct < 1
            and scenario.motor_power_override_kw_fc_demand_on_pct > 0
        ), f"motor_power_override_kw_fc_demand_on_pct {scenario.motor_power_override_kw_fc_demand_on_pct}"
    if a_vehicle is not None and a_vehicle.veh_pt_type == gl.PHEV:
        if scenario.motor_power_override_kw_fc_demand_on_pct != 1:
            assert (
                a_vehicle.kw_demand_fc_on != None
                and np.isnan(a_vehicle.kw_demand_fc_on) != True
            )
    assert (
        scenario.phev_utility_factor_computed == -1
    ), "this should never be populated in input files, only computed if user does not populate phev_utility_factor_override"

    return scenario


def load_design_cycle_from_scenario(
    scenario: Scenario, cyc_file_path: str = gl.OPTIMIZATION_DRIVE_CYCLES
) -> fastsim.cycle.Cycle:
    """
    This helper method loads the design cycle used for mpgge and range determination.
    It can also be used standalone to get cycles not in standard gl.OPTIMIZATION_DRIVE_CYCLES location,
    but still needs cycle name from scenario object, carried in scenario.drive_cycle.
    If the drive cycles are a list of tuples, handle accordingly with eval.

    Args:
        scenario (Scenario): Scenario object for current selection
        cyc_file_path (str, optional): drivecycle input file path. Defaults to gl.OPTIMIZATION_DRIVE_CYCLES.

    Returns:
        range_cyc (fastsim.cycle.Cycle): FASTSim cycle object for current Scenario object
    """
    # determine if scenario.drive_cycle is a simple string path or a list of tuples as a string
    sdc = scenario.drive_cycle
    if "[" in sdc and "]" in sdc and "(" in sdc and ")" in sdc:
        scenario.drive_cycle = ast.literal_eval(sdc)
        range_cyc = []
        for dc_weight in scenario.drive_cycle:
            cycle_file_name = Path(dc_weight[0]).name
            dc = load_design_cycle_from_path(
                cyc_file_path=Path(cyc_file_path) / dc_weight[0]
            )
            dc.name = cycle_file_name
            weight = dc_weight[1]
            range_cyc.append((dc, weight))
    else:
        cycle_file_name = Path(scenario.drive_cycle).name
        range_cyc = load_design_cycle_from_path(
            cyc_file_path=Path(cyc_file_path) / scenario.drive_cycle
        )
        range_cyc.name = cycle_file_name

    return range_cyc


def load_design_cycle_from_path(cyc_file_path: str) -> fastsim.cycle.Cycle:
    """
    This helper method loads the Cycle object from the drivecycle filepath

    Args:
        cyc_file_path (str): drivecycle input file path

    Returns:
        range_cyc (fastsim.cycle.Cycle): FASTSim cycle object for current Scenario object
    """
    if Path(cyc_file_path).exists() == False:
        finalized_path = gl.OPTIMIZATION_DRIVE_CYCLES
        print(f"drive cycle not found in {cyc_file_path} trying {finalized_path}")
    else:
        finalized_path = cyc_file_path
    range_cyc = cycle.Cycle.from_file(finalized_path)
    range_cyc = range_cyc.to_rust()
    return range_cyc


# ---------------------------------- powertrain adjustment methods ---------------------------------- #


def set_test_weight(vehicle: fastsim.vehicle.Vehicle, scenario: Scenario) -> None:
    """
    assign standardized vehicle mass for accel and grade test using GVWR and GVWR Credit

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        scenario (t3co.run_scenario.Scenario): T3CO scenario object
    """
    # June 15,16 confirming that the test weight of vehicle should be GVWRKg + gvwr_credit_kg
    vehicle.veh_override_kg = scenario.gvwr_kg + scenario.gvwr_credit_kg
    vehicle.set_veh_mass()
    assert (
        vehicle.veh_kg > 0
    ), "vehicle weight [kg] cannot be zero, check Scenario values for gvwr_kg and gvwr_credit_kg"


def reset_vehicle_weight(vehicle: fastsim.vehicle.Vehicle) -> None:
    """
    This function resets vehicle mass after loaded weight tests are done for accel and grade

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
    """
    vehicle.veh_override_kg = 0
    vehicle.set_veh_mass()


def limit_cargo_kg_for_moo_hev_bev(
    opt_scenario: Scenario, mooadvancedvehicle: fastsim.vehicle.Vehicle
) -> None:
    """
    This helper method is used within T3COProblem to assign limited cargo capacity based on GVWR + GVWRCredit and optimization vehicle mass for advanced vehicles

    Args:
        opt_scenario (t3co.run_scenario.Scenario): T3CO scenario object
        mooadvancedvehicle (fastsim.vehicle.Vehicle): pymoo optimization vehicle
    """
    # limit cargo to a value <= its original mass, decrease it if vehicle is overweight
    max_allowable_weight_kg = opt_scenario.gvwr_kg + opt_scenario.gvwr_credit_kg
    cargo_limited = max_allowable_weight_kg - (
        mooadvancedvehicle.veh_kg - mooadvancedvehicle.cargo_kg
    )
    cargo_limited = max(cargo_limited, 0)
    # TODO socialize the fact that this next line makes it impossible to add cargo capacity relative to baseline
    # lightweighting and such can improve energy efficiency but not increase cargo
    mooadvancedvehicle.cargo_kg = min(cargo_limited, opt_scenario.originalcargo_kg)
    mooadvancedvehicle.set_veh_mass()


# helper methods to ensure users call proper vehicle initialization methods to adjust vehicle powertrain and mass
def set_max_motor_kw(
    analysis_vehicle: fastsim.vehicle.Vehicle, scenario: Scenario, max_motor_kw: float
) -> None:
    """
    This helper method is used within T3COProblem to set max_motor_kw to optimization vehicle and set kw_demand_fc_on if PHEV

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        scenario (run_scenario.Scenario): T3CO Scenarion object
        max_motor_kw (float): max motor power /kW
    """
    # old comments that may be needed again:
    # Scaling motor and ESS power with ESS capacity results in more reasonable
    # zero-to-sixty response to battery capacity and is generally consistent
    # with how things are done.  We need to firm up the functional form of this,
    # which came from Aaron Brooker for light duty.
    # veh.mc_max_kw = 24.46 * (veh.ess_max_kwh ** (-.475) * veh.ess_max_kwh)
    analysis_vehicle.mc_max_kw = max_motor_kw
    # TODO: for HEV (at least), battery power could be significantly lower than motor power,
    # and the following variable assignment will be pretty far off

    analysis_vehicle.ess_max_kw = (
        analysis_vehicle.mc_max_kw / analysis_vehicle.get_mcPeakEff()
    )

    # PHEV adjustment
    if analysis_vehicle.veh_pt_type == gl.PHEV:
        if scenario.motor_power_override_kw_fc_demand_on_pct != -1:
            analysis_vehicle.kw_demand_fc_on = (
                max_motor_kw * scenario.motor_power_override_kw_fc_demand_on_pct
            )

    analysis_vehicle.set_derived()


def set_max_battery_kwh(
    analysis_vehicle: fastsim.vehicle.Vehicle, max_ess_kwh: float
) -> None:
    """
    This helper method is used within T3COProblem to set max_ess_kwh to optimization vehicle

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        max_ess_kwh (float): max energy storage system energy capacity /kWh
    """
    analysis_vehicle.ess_max_kwh = max_ess_kwh
    analysis_vehicle.set_derived()


def set_max_battery_power_kw(
    analysis_vehicle: fastsim.vehicle.Vehicle, max_ess_kw: float
) -> None:
    """
    This helper method is used within T3COProblem to set max_ess_kwx to optimization vehicle

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        max_ess_kw (float): max energy storage system power /kW
    """
    analysis_vehicle.ess_max_kw = max_ess_kw
    analysis_vehicle.set_derived()


def set_max_fuel_converter_kw(
    analysis_vehicle: fastsim.vehicle.Vehicle, fc_max_out_kw: float
) -> None:
    """
    This helper method is used within T3COProblem to set fc_max_out_kw to optimization vehicle

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        fc_max_out_kw (float): max fuel converter power /kW
    """
    analysis_vehicle.fc_max_kw = fc_max_out_kw
    analysis_vehicle.set_derived()


def set_fuel_store_kwh(
    analysis_vehicle: fastsim.vehicle.Vehicle, fs_kwh: float
) -> None:
    """
    This helper method is used within T3COProblem to set fs_kwh to optimization vehicle

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        fs_kwh (float): fuel storage energy capacity /kWh
    """
    analysis_vehicle.fs_kwh = fs_kwh
    analysis_vehicle.set_derived()


def set_cargo_kg(analysis_vehicle: fastsim.vehicle.Vehicle, cargo_kg):
    """
    This helper method is used within T3COProblem to set cargo_kg to optimization vehicle

    Args:
        analysis_vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        cargo_kg (float): vehicle cargo capacity /kg
    """
    analysis_vehicle.cargo_kg = cargo_kg
    analysis_vehicle.set_veh_mass()


def vehicle_scenario_sweep(
    vehicle: fastsim.vehicle.Vehicle,
    scenario: Scenario,
    range_cyc,
    verbose=False,
    **kwargs,
):
    """
    This function contains helper methods such as get_tco_of_vehicle, check_phev_init_socs, get_accel, and get_gradeability\
    and returns a dictionary of all TCO related outputs

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object for current selection
        scenario (Scenario): Scenario object for current selection
        range_cyc (fastsim.cycle.Cycle): FASTSim cycle object for current scenario
        verbose (bool, optional): if selected, prints out the TCO calculation process. Defaults to False.

    Returns:
        out (dict): output dictionary containing TCO elements
    """
    get_accel = kwargs.get("get_accel", True)
    get_accel_loaded = kwargs.get("get_accel_loaded", True)
    get_gradability = kwargs.get("get_gradability", True)
    write_tsv = kwargs.get("write_tsv", False)

    # run the vehicle through TCO calculations
    if verbose:
        print("Running `tco_analysis.get_tco_of_vehicle`")
    (
        tot_cost_dol,
        discounted_tco_dol,
        oppy_cost_set,
        ownership_costs_df,
        discounted_costs_df,
        mpgge,
        veh_cost_set,
        design_cycle_sdr,
        veh_oper_cost_set,
        veh_opp_cost_set,
        tco_files,
    ) = tco_analysis.get_tco_of_vehicle(
        vehicle, range_cyc, scenario, write_tsv=write_tsv
    )

    # tco_analysis.get_operating_costs(scenario, ownership_costs_df, veh_opp_cost_set)

    vehicle_mass = {
        "glider_kg": vehicle.glider_kg,
        "cargo_kg": vehicle.cargo_kg,
        "transKg": vehicle.trans_kg * vehicle.comp_mass_multiplier,
        "ess_mass_kg": vehicle.ess_mass_kg,
        "mc_mass_kg": vehicle.mc_mass_kg,
        "fc_mass_kg": vehicle.fc_mass_kg,
        "fs_mass_kg": vehicle.fs_mass_kg,
        "veh_kg": vehicle.veh_kg,
        "gliderLb": gl.kg_to_lbs(vehicle.glider_kg),
        "cargoLb": gl.kg_to_lbs(vehicle.cargo_kg),
        "transLb": gl.kg_to_lbs(vehicle.trans_kg * vehicle.comp_mass_multiplier),
        "essMasLb": gl.kg_to_lbs(vehicle.ess_mass_kg),
        "mcMassLb": gl.kg_to_lbs(vehicle.mc_mass_kg),
        "fcMassLb": gl.kg_to_lbs(vehicle.fc_mass_kg),
        "fsMassLb": gl.kg_to_lbs(vehicle.fs_mass_kg),
        "vehLb": gl.kg_to_lbs(vehicle.veh_kg),
    }
    grade_sdr_6 = None
    grade_sdr_125 = None
    accel_sdr = None
    accel_loaded_sdr = None
    zero_to_60 = None
    zero_to_30 = None
    zero_to_60_loaded = None
    zero_to_30_loaded = None
    grade_6_mph_ach = None
    grade_1_25_mph_ach = None
    ess_init_soc_accel = None
    ess_init_soc_grade = None

    # init SOC overrides for grade and accel, for any vehicle that is not a PHEV
    if scenario.ess_init_soc_grade != -1:
        ess_init_soc_grade = scenario.ess_init_soc_grade
    if scenario.ess_init_soc_accel != -1:
        ess_init_soc_accel = scenario.ess_init_soc_accel

    check_phev_init_socs(vehicle, scenario)

    if scenario.soc_norm_init_for_grade_pct != -1:
        ess_init_soc_grade = vehicle.min_soc + (
            scenario.soc_norm_init_for_grade_pct * (vehicle.max_soc - vehicle.min_soc)
        )
    if scenario.soc_norm_init_for_accel_pct != -1:
        ess_init_soc_accel = vehicle.min_soc + (
            scenario.soc_norm_init_for_accel_pct * (vehicle.max_soc - vehicle.min_soc)
        )

    if get_accel:
        if verbose:
            print(f"{gl.SWEEP_PATH.name}:: Running accel.get_accel")
        zero_to_60, zero_to_30, accel_sdr = accel.get_accel(
            vehicle,
            scenario,
            set_weight_to_max_kg=False,
            ess_init_soc=ess_init_soc_accel,
            verbose=verbose,
        )
    if get_accel_loaded:
        if verbose:
            print(f"{gl.SWEEP_PATH.name}:: Running accel.get_accel loaded")
        zero_to_60_loaded, zero_to_30_loaded, accel_loaded_sdr = accel.get_accel(
            vehicle,
            scenario,
            set_weight_to_max_kg=True,
            ess_init_soc=ess_init_soc_accel,
            verbose=verbose,
        )
    if get_gradability:
        if verbose:
            print(f"{gl.SWEEP_PATH.name}:: Running gradeability.get_gradeability")
        (
            grade_6_mph_ach,
            grade_1_25_mph_ach,
            grade_sdr_6,
            grade_sdr_125,
        ) = gradeability.get_gradeability(
            vehicle,
            scenario,
            ess_init_soc=ess_init_soc_grade,
            set_weight_to_max_kg=True,
        )

    range_dict = fueleconomy.get_range_mi(mpgge, vehicle, scenario)

    out = {
        "discounted_costs_df": discounted_costs_df,
        "veh_oper_cost_set": veh_oper_cost_set,
        "veh_opp_cost_set": veh_opp_cost_set,
        "mpgge": mpgge,
        "veh_msrp_set": veh_cost_set,
        "vehicle": vehicle,
        "vehicle_mass": vehicle_mass,
        "zero_to_60": zero_to_60,
        "zero_to_30": zero_to_30,
        "zero_to_60_loaded": zero_to_60_loaded,
        "zero_to_30_loaded": zero_to_30_loaded,
        "grade_6_mph_ach": grade_6_mph_ach,
        "grade_1_25_mph_ach": grade_1_25_mph_ach,
        "scenario": scenario,
        "design_cycle_sim_drive_record": design_cycle_sdr,
        "accel_sim_drive_record": accel_sdr,
        "accel_loaded_sim_drive_record": accel_loaded_sdr,
        "grade_6_sim_drive_record": grade_sdr_6,
        "grade_1p25_sim_drive_record": grade_sdr_125,
        "disc_cost": discounted_tco_dol,
        "opportunity_cost_set": oppy_cost_set,
        "tot_cost": tot_cost_dol,
        "tco_files": tco_files,
    }
    out.update(range_dict)
    return out


def run(
    veh_no: int,
    vocation: str = "blank",
    vehicle_input_path: str = gl.FASTSIM_INPUTS,
    scenario_inputs_path: str = gl.OTHER_INPUTS,
):
    """
    This function runs vehicle_scenario_sweep based on vehicle and scenario objects read from input file paths

    Args:
        veh_no (int): vehicle selection number
        vocation (str, optional): vocation description of selected vehicle. Defaults to "blank".
        vehicle_input_path (str, optional): input file path for vehicle assumptions CSV. Defaults to gl.FASTSIM_INPUTS.
        scenario_inputs_path (str, optional): input file path for scenario assumptions CSV. Defaults to gl.OTHER_INPUTS.

    Returns:
        out (dict): output dictionary containing TCO results
    """

    # set up tco results directories for the vocation-scenario
    gl.vocation_scenario = vocation
    gl.set_tco_intermediates()
    gl.set_tco_results()

    # load the generated file of vehicles, drive cycles, and tech targets
    vehicle = get_vehicle(veh_no, vehicle_input_path)
    scenario, range_cyc = get_scenario_and_cycle(veh_no, scenario_inputs_path)

    out = vehicle_scenario_sweep(vehicle, scenario, range_cyc)

    return out


def rerun(vehicle: fastsim.vehicle.Vehicle, vocation: str, scenario: Scenario):
    """
    This function runs vehicle_scenario_sweep when given the vehicle and scenario objects

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object
        vocation (str): vocation description
        scenario (Scenario): Scenario object

    Returns:
        out (dict): output dictionary containing TCO outputs
    """
    # set up tco results directories for the vocation-scenario
    gl.vocation_scenario = vocation
    gl.set_tco_intermediates()
    gl.set_tco_results()

    range_cyc = load_design_cycle_from_scenario(scenario)

    out = vehicle_scenario_sweep(vehicle, scenario, range_cyc)

    return out



#!/usr/bin/env python
# coding: utf-8

# %%
from typing import List, Tuple
import fastsim
import pandas as pd
import numpy as np
import csv
import time
from time import gmtime, strftime
from pathlib import Path
import argparse
import ast
import logging
import re
import os
import pymoo
import pymoo.core
import pymoo.core.result
from t3co.run import Global as gl, run_scenario
from t3co.moopack import moo
from t3co.run import run_scenario as rs
from t3co.objectives import fueleconomy as fe

REPORT_COLS = {
    "selection": "",
    "scenario_name": "",
    "veh_year": "",
    "veh_pt_type": "",
    "pareto_front_number": "",
    "run_time_[s]": "",
    "algorithm": "",
    "n_gen": "",
    "fvals_over_gens": "",
    "design_cyc_trace_miss_dist_frac": "",
    "design_cyc_trace_miss_time_frac": "",
    "design_cyc_trace_miss_speed_mps": "",
    "design_cycle_EA_err": "",
    "accel_EA_err": "",
    "accel_loaded_EA_err": "",
    "grade_6_EA_err": "",
    "grade_1p25_EA_err": "",
    "final_cda_pct": "",
    "final_eng_eff_pct": "",
    "final_ltwt_pct": "",
    "final_max_motor_kw": "",
    "final_battery_kwh": "",
    "final_max_fc_kw": "",
    "final_fs_kwh": "",
    "range_ach_mi": "",
    "target_range_mi": "",
    "delta_range_mi": "",
    "min_speed_at_6pct_grade_in_5min_ach_mph": "",
    "target_min_speed_at_6pct_grade_in_5min_mph": "",
    "delta_min_speed_at_6pct_grade_in_5min_mph": "",
    "min_speed_at_1p25pct_grade_in_5min_ach_mph": "",
    "target_min_speed_at_1p25pct_grade_in_5min_mph": "",
    "delta_min_speed_at_1p25pct_grade_in_5min_mph": "",
    "max_time_0_to_60mph_at_gvwr_ach_s": "",
    "target_max_time_0_to_60mph_at_gvwr_s": "",
    "delta_max_time_0_to_60mph_at_gvwr_s": "",
    "max_time_0_to_30mph_at_gvwr_ach_s": "",
    "target_max_time_0_to_30mph_at_gvwr_s": "",
    "delta_max_time_0_to_30mph_at_gvwr_s": "",
    "glider_cost_dol": "",
    "fuel_converter_cost_dol": "",
    "fuel_storage_cost_dol": "",
    "motor_control_power_elecs_cost_dol": "",
    "plug_cost_dol": "",
    "battery_cost_dol": "",
    "purchase_tax_dol": "",
    "msrp_total_dol": "",
    "total_fuel_cost_dol": "",
    "total_maintenance_cost_dol": "",
    "discounted_tco_dol": "",
}


# knob mins that are too high lead to failures in optimization
# and intelligently set knob mins don't really aid optimizer speed
# best not to get too cute with it, will only waste time if you do it wrong
KNOB_MIN = 1


def deug_traces(
    vehicle: fastsim.vehicle.Vehicle,
    cycles: List[fastsim.cycle.Cycle],
    scenario: run_scenario.Scenario,
) -> None:
    """
    This function gets a diagnostic trace of get_mpgge

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim Vehicle object
        cycles (List[fastsim.cycle.Cycle]): List of FASTSim drivecycle objects
        scenario (run_scenario.Scenario): Scenario object
    """
    mpgge_comp, sim_drives, mpgges = fe.get_mpgge(
        cycles, vehicle, scenario, diganostic=True
    )
    # save the sim_drives to file somehow


def save_tco_files(
    tco_files: dict, resdir: str, scenario_name: str, sel: str, ts: str
) -> None:
    """
    This function saves the intermediary files as tsv

    Args:
        tco_files (dict): Contains all TCO calculation dataframes
        resdir (str): result directory strong
        scenario_name (str): scenario name
        sel (str): selection(s)
        ts (str): timestring
    """
    scenario_name = re.sub("s", "-", scenario_name)
    scenario_name = re.sub("[^a-zA-Z0-9 \n.]", "", scenario_name)
    tco_files_path = resdir / f"tco_files_{ts}" / f"{sel}_{scenario_name}"
    tco_files_path.mkdir(parents=True, exist_ok=True)

    veh_eff_df = tco_files["veh_eff_df"]
    veh_exp_df = tco_files["veh_exp_df"]
    veh_txp_df = tco_files["veh_txp_df"]
    veh_shr_df = tco_files["veh_shr_df"]
    veh_fxp_df = tco_files["veh_fxp_df"]
    ann_trv_df = tco_files["ann_trv_df"]
    reg_sls_df = tco_files["reg_sls_df"]
    survivl_df = tco_files["survivl_df"]
    veh_spt_df = tco_files["veh_spt_df"]
    stock = tco_files["stock"]
    ownership_costs_df = tco_files["ownership_costs_df"]
    discounted_costs_df = tco_files["discounted_costs_df"]

    veh_eff_df.to_csv(tco_files_path / gl.FUEL_EFF_TSV, sep="\t", index=False)
    veh_exp_df.to_csv(tco_files_path / gl.VEH_EXP_TSV, index=False, sep="\t")
    veh_txp_df.to_csv(tco_files_path / gl.TRAVEL_EXP_TSV, index=False, sep="\t")
    veh_shr_df.to_csv(tco_files_path / gl.MARKET_SHARE_TSV, index=False, sep="\t")
    veh_fxp_df.to_csv(tco_files_path / gl.FUEL_EXPENSE_TSV, index=False, sep="\t")
    ann_trv_df.to_csv(tco_files_path / gl.ANN_TRAVEL_TSV, index=False, sep="\t")
    reg_sls_df.to_csv(tco_files_path / gl.REGIONAL_SALES_TSV, index=False, sep="\t")
    survivl_df.to_csv(tco_files_path / gl.SURVIVAL_TSV, index=False, sep="\t")
    veh_spt_df.to_csv(tco_files_path / gl.FUEL_SPLIT_TSV, index=False, sep="\t")

    stock.to_csv(tco_files_path / "stock_v01.tsv", sep="\t", index=False)
    # emissions.to_csv(                  tco_files_path / 'emissions_v01.tsv', sep='\t', index=False)
    ownership_costs_df.to_csv(
        tco_files_path / "ownership-costs_v01.tsv", sep="\t", index=False
    )
    discounted_costs_df.to_csv(
        tco_files_path / "discounted-ownership-costs_v01.tsv", sep="\t", index=False
    )


def get_knobs_bounds_curves(
    selection: int,
    vpttype: str,
    sdf: pd.DataFrame,
    lw_imp_curves: pd.DataFrame,
    aero_drag_imp_curves: pd.DataFrame,
    eng_eff_curves: pd.DataFrame,
) -> Tuple[dict, dict]:
    """
    This function fetches the knobs and constraints for running the optimization for a given selection

    Args:
        selection (int): selection number
        vpttype (str): vehicle powertrain type = veh_pt_type
        sdf (pd.DataFrame): scenario dataframe
        lw_imp_curves (pd.DataFrame): light weighting curve dataframe
        aero_drag_imp_curves (pd.DataFrame): aero drag curve dataframe
        eng_eff_curves (pd.DataFrame): engine efficiency curve dataframe

    Returns:
        knobs_bounds (dict): dict of knobs and bounds
        curves (dict): dict of lw, aero, and engine efficiency curve parameters
    """
    lw_imp_curves_file = lw_imp_curves.set_index("name")
    aero_drag_imp_curves_file = aero_drag_imp_curves.set_index("name")
    eng_eff_curves_file = eng_eff_curves.set_index("name")
    sd = dict(sdf.loc[selection, :])
    curves = {}

    knobs_bounds = {
        "ess_max_kwh": [sd.get("knob_min_ess_kwh"), sd.get("knob_max_ess_kwh")],
        "mc_max_kw": [sd.get("knob_min_motor_kw"), sd.get("knob_max_motor_kw")],
        "fcMaxOutKw": [sd.get("knob_min_fc_kw"), sd.get("knob_max_fc_kw")],
        "fs_kwh": [sd.get("knob_min_fs_kwh"), sd.get("knob_max_fs_kwh")],
    }

    if (
        "lw_imp_curve_sel" in sd
        and sd["lw_imp_curve_sel"]
        and not pd.isnull(sd["lw_imp_curve_sel"])
    ):
        lw_curve_selection = sd["lw_imp_curve_sel"]
        ltwt_cost_curve = lw_imp_curves_file.loc["ltwt_cost", lw_curve_selection]
        ltwt_pct_curve = lw_imp_curves_file.loc["ltwt_pct", lw_curve_selection]
        wt_delta_perc_knob_max = float(
            lw_imp_curves_file.loc["wt_delta_perc_knob_max", lw_curve_selection]
        )
        wt_delta_perc_knob_min = float(
            lw_imp_curves_file.loc["wt_delta_perc_knob_min", lw_curve_selection]
        )
        assert (
            wt_delta_perc_knob_max <= 1
        ), f"input invalid, value for wt_delta_perc_knob_max must decimal form, got percentage point value as {wt_delta_perc_knob_max}"

        knobs_bounds["wt_delta_perc"] = (wt_delta_perc_knob_min, wt_delta_perc_knob_max)
        curves.update(
            {
                "ltwt_delta_percs": np.array(ast.literal_eval(ltwt_pct_curve)),
                "ltwt_dol_per_kg_costs": np.array(ast.literal_eval(ltwt_cost_curve)),
            }
        )

    if (
        "aero_drag_imp_curve_sel" in sd
        and sd["aero_drag_imp_curve_sel"]
        and not pd.isnull(sd["aero_drag_imp_curve_sel"])
    ):
        cda_curve_selection = sd["aero_drag_imp_curve_sel"]
        CdA_perc_imp_at_which_wt_penalty_maxes_out = float(
            aero_drag_imp_curves_file.loc[
                "CdA_perc_imp_at_which_wt_penalty_maxes_out", cda_curve_selection
            ]
        )
        CdA_perc_imp_knob_max = float(
            aero_drag_imp_curves_file.loc["CdA_perc_imp_knob_max", cda_curve_selection]
        )
        CdA_perc_imp_knob_min = float(
            aero_drag_imp_curves_file.loc["CdA_perc_imp_knob_min", cda_curve_selection]
        )
        cost_a = float(aero_drag_imp_curves_file.loc["cost_a", cda_curve_selection])
        cost_b = float(aero_drag_imp_curves_file.loc["cost_b", cda_curve_selection])
        mass_a = float(aero_drag_imp_curves_file.loc["mass_a", cda_curve_selection])
        mass_b = float(aero_drag_imp_curves_file.loc["mass_b", cda_curve_selection])
        assert (
            CdA_perc_imp_knob_max <= 1
        ), f"input invalid, value for CdA_perc_imp_knob_max must decimal form, got percentage point value as {CdA_perc_imp_knob_max}"
        assert (
            CdA_perc_imp_at_which_wt_penalty_maxes_out <= 1
        ), f"input invalid, value for CdA_perc_imp_at_which_wt_penalty_maxes_out must decimal form, got percentage point value as {CdA_perc_imp_at_which_wt_penalty_maxes_out}"

        knobs_bounds["CdA_perc_imp"] = (CdA_perc_imp_knob_min, CdA_perc_imp_knob_max)
        curves.update(
            {
                "cda_cost_coeff_a": cost_a,
                "cda_cost_coeff_b": cost_b,
                "cda_mass_coeff_a": mass_a,
                "cda_mass_coeff_b": mass_b,
                "cda_perc_imp_at_which_wt_penalty_maxes_out": CdA_perc_imp_at_which_wt_penalty_maxes_out,
            }
        )

    if (
        "eng_eff_imp_curve_sel" in sd
        and sd["eng_eff_imp_curve_sel"]
        and not pd.isnull(sd["eng_eff_imp_curve_sel"])
        and vpttype != gl.BEV
    ):
        # TODO, FCEV should not get eng imp curve parameter
        eng_imp_curve_selection = sd["eng_eff_imp_curve_sel"]
        fc_peak_eff_knob_min = float(
            eng_eff_curves_file.loc["fc_peak_eff_knob_min", eng_imp_curve_selection]
        )
        fc_peak_eff_knob_max = float(
            eng_eff_curves_file.loc["fc_peak_eff_knob_max", eng_imp_curve_selection]
        )
        eng_pctpt = eng_eff_curves_file.loc["eng_pctpt", eng_imp_curve_selection]
        eng_cost = eng_eff_curves_file.loc["eng_cost", eng_imp_curve_selection]
        assert (
            fc_peak_eff_knob_max <= 1
        ), f"input invalid, value for fc_peak_eff_knob_max must decimal form, got percentage point value as {fc_peak_eff_knob_max}"

        curves.update(
            {
                "fc_eff_array": np.array(ast.literal_eval(eng_pctpt)),
                "fc_cost_coeff_array": np.array(ast.literal_eval(eng_cost)),
            }
        )

        knobs_bounds["fc_peak_eff"] = (fc_peak_eff_knob_min, fc_peak_eff_knob_max)

    # validate knobs bounds and remove inactive knobs (user left blank or NA)
    knobs = list(knobs_bounds.keys())
    print("knobs and bounds")
    for k in knobs:
        print(k, knobs_bounds[k][0], knobs_bounds[k][1])
        nans = np.isnan(knobs_bounds[k][0]) or np.isnan(knobs_bounds[k][1])
        if nans or knobs_bounds[k][0] is None or knobs_bounds[k][1] is None:
            del knobs_bounds[k]
    return knobs_bounds, curves


def get_objectives_constraints(
    selection: int, sdf: pd.DataFrame, verbose: bool = True
) -> Tuple[list, list]:
    """
    This function appends to list of necessary variables based on the constraints and objectives selected

    Args:
        selection (int): selection number
        sdf (DataFrame): scenario dataframe
        verbose (bool, optional): if selected, function will print objectives and constraints. Defaults to True.

    Returns:
        objectives (list): list of selected objective variables
        constraints (list): list of selected constraint variables
    """
    sd = dict(sdf.loc[selection, :])
    objectives = []
    constraints = []
    # constraints
    if sd.get("constraint_range") == True:
        constraints.append(moo.RANGE)
    if sd.get("constraint_accel") == True:
        constraints.extend([moo.ACCEL30, moo.ACCEL60])
    if sd.get("constraint_grade") == True:
        constraints.extend([moo.GRADE125, moo.GRADE6])
    if sd.get("constraint_c_rate") == True:
        constraints.append(moo.C_RATE)
    if sd.get("constraint_trace_miss_dist_percent_on") == True:
        constraints.append(moo.TRACE_MISS_DIST_PERCENT)
    if sd.get("constraint_phev_minimize_fuel_use_on") == True:
        constraints.append(moo.PHEV_MINIMIZE_FUEL_USE_CONSTRAINT)
    # objectives
    if sd.get("objective_phev_minimize_fuel_use") == True:
        objectives.append(moo.PHEV_MINIMIZE_FUEL_USE_OBJECTIVE)
    if sd.get("objective_tco") == True:
        objectives.append(moo.TCO)

    if verbose:
        print("objectives:", objectives)
        print("constraints:", constraints)

    return objectives, constraints


def run_moo(
    sel: int,
    sdf: pd.DataFrame,
    optpt: str,
    algo: str,
    skip_opt: bool,
    pop_size: float,
    n_max_gen: int,
    n_last: int,
    nth_gen: int,
    x_tol: float,
    verbose: bool,
    f_tol: float,
    resdir: str,
    lw_imp_curves: pd.DataFrame,
    aero_drag_imp_curves: pd.DataFrame,
    eng_eff_imp_curves: pd.DataFrame,
    config: run_scenario.Scenario,
    **kwargs,
) -> Tuple[pymoo.core.result.Result, moo.T3COProblem, bool]:
    """
    This function calls get_objectives_constraints and get_knobs_bounds_curves, and then calls run_optimization to perform the multiobjective optimization

    Args:
        sel (int): selection number
        sdf (DataFrame): scenario dataframe
        optpt (str): vehicle powertrain type
        algo (str): algorithm name
        skip_opt (bool): skip optimization boolean
        pop_size (int): population size for optimization
        n_max_gen (int): maximum number of generations for optimization
        n_last (int): number of generations to look back for termination
        nth_gen (int): number of generations to evaluate if convergence occurs
        x_tol (float): tolerance in parameter space
        verbose (book): if selected, function prints the optimization process
        f_tol (float): tolerance in objective space
        resdir (str): results directory
        lw_imp_curves (DataFrame): light weighting curves dataframe
        aero_drag_imp_curves (DataFrame): aero drag curves dataframe
        eng_eff_imp_curves (DataFrame): engine efficiency curve dataframe
        config (Config): Config class object

    Returns:
        moo_results (pymoo.core.result.Result): optimization results object
        moo_problem (T3COProblem): minimization problem that calculates TCO
        moo_code (bool): Error message
    """
    objectives, constraints = get_objectives_constraints(sel, sdf)

    knobs_bounds, curve_settings = get_knobs_bounds_curves(
        sel, optpt, sdf, lw_imp_curves, aero_drag_imp_curves, eng_eff_imp_curves
    )

    # moo_reults has res.X, res.F np arrays of opt params & objective space resutls, respectively
    moo_results, moo_problem, moo_code = moo.run_optimization(
        pop_size,
        n_max_gen,
        knobs_bounds,
        sel,
        optimize_pt=optpt,
        skip_optimization=skip_opt,
        obj_list=objectives,
        constr_list=constraints,
        verbose=verbose,
        n_last=n_last,
        algo=algo,
        nth_gen=nth_gen,
        x_tol=x_tol,
        f_tol=f_tol,
        config=config,
        **curve_settings,
        **kwargs,
    )
    # return input_vehicle, report_vehicle, report_scenario, out_dict
    return moo_results, moo_problem, moo_code


def check_input_files(df: pd.DataFrame, filetype: str, filepath: str) -> None:
    """
    This function contains assert statements that make sure input vehicle and scenario dataframes do not contain numm rows

    Args:
        df (DataFrame): vehicle or scenario dataframe
        filetype (str): 'vehicle' or 'scenario'
        filepath (str): filepath of the vehicle or scenario input files
    """
    blank_lines = [i for i in df.index.isnull().cumsum() if i != 0]
    assert (
        df.index.isnull().any() == False
    ), f"\n\n{filetype} file selection column cannot have blank values\nlines: {blank_lines}\npath: {filepath}\n\n\n\n"  # noqa: E712
    assert (
        df["scenario_name"].isnull().any() == False
    ), f"\n\n{filetype} file scenario_name column cannot have blank values\nlines: {blank_lines}\npath: {filepath}\n\n\n\n"  # noqa: E712


def run_vehicle_scenarios(
    vehicles: str,
    scenarios: str,
    eng_eff_imp_curves_p: str,
    lw_imp_curves_p: str,
    aero_drag_imp_curves_p: str,
    config: run_scenario.Config,
    **kwargs,
) -> None:
    """
    This is the main function that runs T3CO for all the selections input

    Args:
        vehicles (str): path of vehicle input file
        scenarios (str): path of scenarios input file
        eng_eff_imp_curves_p (str): path of engine efficiency curve file
        lw_imp_curves_p (str): path of light weighting curve file
        aero_drag_imp_curves_p (str): path of aero drag curve file
        config (Config): Config object containing analysis attributes and scenario attribute overrides

    Raises:
        Exception: input validation error
        Exception: optimization error

    """
    vdf = pd.read_csv(vehicles, index_col="selection", skip_blank_lines=True)
    sdf = pd.read_csv(scenarios, index_col="selection", skip_blank_lines=True)
    check_input_files(vdf, "vehicles", vehicles)
    check_input_files(sdf, "scenario", scenarios)
    eng_eff_imp_curves = pd.read_csv(eng_eff_imp_curves_p)
    lw_imp_curves = pd.read_csv(lw_imp_curves_p)
    aero_drag_imp_curves = pd.read_csv(aero_drag_imp_curves_p)

    # optimization scenario parameters where we get our baseline vehicle and scenario
    gl.FASTSIM_INPUTS = vehicles
    gl.OTHER_INPUTS = scenarios
    print("vehicle src:", vehicles)
    print("scenario src:", scenarios)

    for scen_key in rs.Scenario.__dict__["__annotations__"].keys():
        REPORT_COLS.update({"scenario_" + scen_key: ""})

    df = pd.read_csv(gl.FASTSIM_INPUTS)

    for vc in df.columns.tolist():
        REPORT_COLS.update({"optimized_vehicle_value_" + vc: ""})

    for vc in df.columns.tolist():
        REPORT_COLS.update({"input_vehicle_value_" + vc: ""})

    # optimizer and sweep tuning
    report_kwargs = dict(kwargs)
    n_max_gen = kwargs.pop(
        "n_max_gen", 100
    )  # max number of generations if convergence not met
    pop_size = kwargs.pop("pop_size", 30)  # population of each generation
    nth_gen = kwargs.pop(
        "nth_gen", 1
    )  # period of generations in which to evaluate if convergence happens
    algorithms = kwargs.pop("algo")
    n_last = kwargs.pop(
        "n_last", 5
    )  # number of generations to look back for establishing convergence
    x_tol = kwargs.pop("x_tol", 0.5)  # parameter space tolerance
    f_tol = kwargs.pop("f_tol", 3.0)  # objective space tolerance
    verbose = kwargs.pop("verbose", False)
    look_for = kwargs.pop("look_for", [""])
    assert isinstance(
        look_for, list
    ), "look_for should have been input or cast as a list at this point"
    selections = kwargs.pop(
        "selections", -1
    )  # list of integers representing scenario selections, ex: [34,35,36]
    exclude = kwargs.pop("exclude", [">{-<>-}<"])
    dir_mark = kwargs.pop("dir_mark", "")
    file_mark = kwargs.pop("file_mark", "")
    skip_save_veh = kwargs.pop("skip_save_veh")
    skip_all_opt = kwargs.pop("skip_all_opt")
    do_input_validation = kwargs.pop("do_input_validation")
    write_tsv = kwargs.pop("do_input_validation", False)

    if not file_mark.endswith("_"):
        file_mark += "_"

    # results dir setup
    ts = strftime("%Y-%m-%d_%H-%M-%S", gmtime())
    RES_FILE = f"{file_mark}results_{ts}_sel_{str(selections).strip('[]').replace(' ','').replace(',','-')}.csv".strip(
        "_"
    )

    if selections == -1:
        selections = range(len(vdf))

    if args.dst_dir is None and config.dst_dir is None:
        resdir = Path(os.path.abspath(__file__)).parents[1] / f"results{dir_mark}"
    elif args.dst_dir is not None:
        resdir = Path(args.dst_dir)
    else:
        resdir = Path(config.dst_dir)

    if not resdir.exists():
        resdir.mkdir(parents=True)

    with open(str(resdir / RES_FILE), "a", newline="") as f:
        print("writing to ", resdir / RES_FILE)
        writer = csv.writer(f)
        writer.writerow(REPORT_COLS.keys())

    loggingfname = Path(
        str(Path(resdir / f"{file_mark}sweep_error_log_{ts}.log")).strip("_")
    )
    logging.basicConfig(
        filename=loggingfname,
        level=logging.INFO,
        filemode="a",
        datefmt="%Y-%m-%d %H:%M:%S",
        format="%(asctime)s %(levelname)-8s %(message)s",
        force=True,
    )
    logging.info(f"kwargs {report_kwargs}")

    # list of report dataframes to write final output at each iteration
    reports = []

    def input_validation(
        sel: float, optpt: str, algo: str, config: run_scenario.Scenario
    ):
        """
        This function obtains the vehicle, scenario, and cycle object for a given selection and runs optimization to validate inputs

        Args:
            sel (float): selection number
            optpt (str): vehicle powertrain type
            algo (str): algorithm name
            config (Config): Config object

        Returns:
            None: None
        """
        v = rs.get_vehicle(
            sel,
            veh_input_path=gl.FASTSIM_INPUTS,
        )
        s, c = rs.get_scenario_and_cycle(
            sel, gl.OTHER_INPUTS, a_vehicle=v, config=config
        )

        rs.check_phev_init_socs(v, s)

        knobs_bounds, curve_settings = get_knobs_bounds_curves(
            sel, optpt, sdf, lw_imp_curves, aero_drag_imp_curves, eng_eff_imp_curves
        )
        objectives, constraints = get_objectives_constraints(sel, sdf, verbose=False)

        _ = moo.run_optimization(
            pop_size,
            n_max_gen,
            knobs_bounds,
            sel,
            optimize_pt=optpt,
            skip_optimization=True,
            obj_list=objectives,
            constr_list=constraints,
            verbose=verbose,
            n_last=n_last,
            algo=algo,
            nth_gen=nth_gen,
            x_tol=x_tol,
            f_tol=f_tol,
            config=config,
            **curve_settings,
            **kwargs,
        )
        return None

    def optimize(
        sel: float,
        scenario_name: str,
        optpt: str,
        algo: str,
        skip_opt: bool,
        write_tsv: bool = False,
    ) -> None:
        """
        This function runs the optimization for a given selection if skip_opt = False

        Args:
            sel (float): selection number
            scenario_name (str): scenario name
            optpt (str): vehicle powertrain type
            algo (str): algorithm name
            skip_opt (bool): skip optimization
            write_tsv (bool, optional): if selected, intermediary dataframes are saved as tsv files. Defaults to False.
        """
        # optimizing is the inverse of skip_opt, just easier bool var name and value to work with

        print(
            f"Running selection {sel} for scenario {scenario_name} - skip opt = {skip_opt} -algo = {algo}"
        )

        ti = time.time()
        sel = float(sel)

        gl.vocation_scenario = scenario_name
        if not skip_opt:
            moo_results, moo_problem, moo_code = run_moo(
                sel,
                sdf,
                optpt,
                algo,
                skip_opt,
                pop_size,
                n_max_gen,
                n_last,
                nth_gen,
                x_tol,
                verbose,
                f_tol,
                resdir,
                lw_imp_curves,
                aero_drag_imp_curves,
                eng_eff_imp_curves,
                config,
                **kwargs,
            )
            num_results = 1
            if moo_code == moo.OPTIMIZATION_SUCCEEDED:
                if moo_results.X.ndim > 1:
                    num_results = moo_results.X.shape[0]
                    # ensure we're only showing unique solution set solutions
                    moo_results.X = np.array(
                        [list(x) for x in set(tuple(x) for x in moo_results.X)]
                    )
            if moo_problem is not None:
                input_vehicle = moo_problem.moobasevehicle
                report_vehicle = moo_problem.mooadvancedvehicle
                report_scenario = moo_problem.opt_scenario
            else:
                input_vehicle = rs.get_vehicle(sel, veh_input_path=gl.FASTSIM_INPUTS)
                report_vehicle = None
                report_scenario, design_cycle = rs.get_scenario_and_cycle(
                    sel, gl.OTHER_INPUTS, a_vehicle=input_vehicle, config=config
                )

            # update records, moo_problem can always be returned unless an exception is thrown above
            if moo_problem is not None:
                moo_problem.compile_reporting_vars()
                if moo_problem.reporting_vars is not None:
                    opt_vars_f_name = (
                        f"{file_mark}_{algo}_var_record_selection_{sel}.csv".strip("_")
                    )
                    moo_problem.reporting_vars.to_csv(resdir / opt_vars_f_name)

        elif skip_opt == True:
            # TODO, is moo_problem.moobasevehicle really the right vehicle here?
            num_results = 1
            moo_code = "NA"
            input_vehicle = rs.get_vehicle(sel, veh_input_path=gl.FASTSIM_INPUTS)
            report_scenario, design_cycle = rs.get_scenario_and_cycle(
                sel, gl.OTHER_INPUTS, a_vehicle=input_vehicle, config=config
            )

            outdict = rs.vehicle_scenario_sweep(
                input_vehicle, report_scenario, design_cycle, write_tsv=write_tsv
            )

        # iterate thru all results from run, num_results can singleton [1] from analysis-only runs
        # or an array or a list of arrays from a parato front result
        # res.X holds results of optimization
        # successful optimization could be 1D (one solution)
        # [1.0, 2.0, 3.0, 4.0, 5.0]
        # or nD, multiple solutions on a pareto front or solution set
        # [[1.0, 2.0, 3.0, 4.0, 5.0],
        # [ 6.0, 7.0, 8.0, 9.0, 10.0],
        # [ 11.0,12.0,13.0,14.0,15.0]]
        for i in range(0, num_results):
            full_report = True

            report_i = {k: "" for k in REPORT_COLS.keys()}

            # important, record sets information: the input vehicle, the scenario, and the optimized vehicle result, if applicable
            for config_key in rs.Config.__dict__["__annotations__"].keys():
                report_i["config_" + config_key] = config.__getattribute__(config_key)

            for scen_key in rs.Scenario.__dict__["__annotations__"].keys():
                report_i["scenario_" + scen_key] = report_scenario.__getattribute__(
                    scen_key
                )
            for v_input_k in input_vehicle.__dict__.keys():
                if "value_props" not in v_input_k:
                    report_i["input_vehicle_value_" + v_input_k] = (
                        input_vehicle.__getattribute__(v_input_k)
                    )
                    # we want place-holder blank values for optimization columns even if we're not optimizing
                    report_i["optimized_vehicle_value_" + v_input_k] = None

            # remove value props object
            if "optimized_vehicle_value_props" in report_i:
                del report_i["optimized_vehicle_value_props"]
            if "input_vehicle_value_props" in report_i:
                del report_i["input_vehicle_value_props"]

            opt_time = round(time.time() - ti)

            report_i["selection"] = sel

            report_i["scenario_name"] = scenario_name
            try:
                report_i["veh_year"] = vdf.loc[sel, "veh_year"]
            except KeyError:
                report_i["veh_year"] = np.nan
            report_i["veh_pt_type"] = vdf.loc[sel, "veh_pt_type"]

            report_i["run_time_[s]"] = opt_time

            report_i["algorithm"] = algo
            n_gens_used = 0
            if not skip_opt:
                if moo_code in [
                    moo.EXCEPTION_THROWN,
                    moo.OPTIMIZATION_FAILED_TO_CONVERGE,
                ]:
                    if moo_code == moo.EXCEPTION_THROWN:
                        report_i["n_gen"] = (
                            "Code Exception thrown"  # TODO, get stacktrace information and add to this
                        )
                    elif moo_code == moo.OPTIMIZATION_FAILED_TO_CONVERGE:
                        report_i["n_gen"] = "Optimization Failed to converge"
                    report_i = {k: str(v) for k, v in report_i.items()}
                    reports.append(report_i)
                    result = report_i["n_gen"]
                    logging.info(f"scenario {sel} failed. {result}")
                    full_report = False

                elif moo_code == moo.OPTIMIZATION_SUCCEEDED:
                    report_i["opt_result_number"] = i + 1
                    if moo_results.X.ndim == 1:
                        x = moo_results.X
                    else:
                        x = moo_results.X[i, :]

                    outdict = moo_problem.get_tco_from_moo_advanced_result(x)

                    # Save resulting vehicle model as YAML file
                    if not skip_save_veh:
                        sim_drives = outdict["design_cycle_sim_drive_record"]
                        for sd in sim_drives:
                            # sd_name = sd.name
                            # sd = sd.to_rust()
                            sd_file_path = (
                                resdir
                                / "sim_drives"
                                / f"{file_mark}sim_drive_result_{int(sel)}.yaml".strip(
                                    "_"
                                )
                            )
                            sd_file_path.parent.mkdir(parents=True, exist_ok=True)
                            sd.to_file(str(sd_file_path))

                        veh_result = (
                            report_vehicle.to_rust()
                            if not skip_opt
                            else input_vehicle.to_rust()
                        )
                        veh_filepath = (
                            resdir
                            / "vehicles"
                            / f"{file_mark}vehicle_result_{int(sel)}.yaml".strip("_")
                        )
                        veh_filepath.parent.mkdir(parents=True, exist_ok=True)
                        veh_result.to_file(str(veh_filepath))

                    x_dixt = {
                        knob: x[moo_problem.knobs.index(knob)]
                        for knob in moo_problem.knobs
                    }
                    report_i["final_cda_pct"] = x_dixt.get(moo.KNOB_CDA)
                    report_i["final_eng_eff_pct"] = x_dixt.get(moo.KNOB_FCPEAKEFF)
                    report_i["final_ltwt_pct"] = x_dixt.get(moo.KNOB_WTDELTAPERC)
                    report_i["final_max_motor_kw"] = x_dixt.get(moo.KNOB_mc_max_kw)
                    report_i["final_battery_kwh"] = x_dixt.get(moo.KNOB_ess_max_kwh)
                    report_i["final_max_fc_kw"] = x_dixt.get(moo.KNOB_FCMAXKW)
                    report_i["final_fs_kwh"] = x_dixt.get(moo.KNOB_fs_kwh)
                    for v_input_k in report_vehicle.__dict__.keys():
                        if "value_props" not in v_input_k:
                            report_i["optimized_vehicle_value_" + v_input_k] = (
                                report_vehicle.__getattribute__(v_input_k)
                            )

                    n_gens_used = moo_results.history[-1].n_gen
                    report_i["fvals_over_gens"] = [
                        e.opt.get("F")[0] for e in moo_results.history
                    ]
                    report_i["n_gen"] = n_gens_used
                    report_i["max_n_gen"] = n_max_gen
            if full_report:
                (
                    tot_cost,
                    disc_cost,
                    oppy_cost_set,
                    discounted_costs_df,
                    mpgge,
                    veh_cost_set,
                    veh_oper_cost_set,
                    veh_opp_cost_set,
                ) = (
                    outdict["tot_cost"],
                    outdict["disc_cost"],
                    outdict["opportunity_cost_set"],
                    outdict["discounted_costs_df"],
                    outdict["mpgge"],
                    outdict["veh_msrp_set"],
                    outdict["veh_oper_cost_set"],
                    outdict["veh_opp_cost_set"],
                )

                print(
                    f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} opt time [s]",
                    opt_time,
                )
                print(
                    f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} total cost",
                    tot_cost,
                )
                print(f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} mpgge", mpgge)
                print(
                    f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} MSRP breakdown",
                    veh_cost_set,
                )
                print(
                    f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} Operating Costs breakdown",
                    veh_oper_cost_set,
                )
                print(
                    f"selection {sel} {gl.PT_TYPES_NUM_TO_STR[optpt]} Opportunity costs breakdown",
                    veh_opp_cost_set,
                )

                disc_cost_agg = discounted_costs_df.groupby("Category").sum(
                    numeric_only=True
                )
                report_i["range_ach_mi"] = outdict["primary_fuel_range_mi"]
                report_i["target_range_mi"] = report_scenario.target_range_mi
                report_i["delta_range_mi"] = (
                    outdict["primary_fuel_range_mi"] - report_scenario.target_range_mi
                )

                report_i["min_speed_at_6pct_grade_in_5min_ach_mph"] = outdict[
                    "grade_6_mph_ach"
                ]
                report_i["target_min_speed_at_6pct_grade_in_5min_mph"] = (
                    report_scenario.min_speed_at_6pct_grade_in_5min_mph
                )
                report_i["delta_min_speed_at_6pct_grade_in_5min_mph"] = (
                    outdict["grade_6_mph_ach"]
                    - report_scenario.min_speed_at_6pct_grade_in_5min_mph
                )

                report_i["min_speed_at_1p25pct_grade_in_5min_ach_mph"] = outdict[
                    "grade_1_25_mph_ach"
                ]
                report_i["target_min_speed_at_1p25pct_grade_in_5min_mph"] = (
                    report_scenario.min_speed_at_1p25pct_grade_in_5min_mph
                )
                report_i["delta_min_speed_at_1p25pct_grade_in_5min_mph"] = (
                    outdict["grade_1_25_mph_ach"]
                    - report_scenario.min_speed_at_1p25pct_grade_in_5min_mph
                )

                report_i["max_time_0_to_60mph_at_gvwr_ach_s"] = outdict[
                    "zero_to_60_loaded"
                ]
                report_i["target_max_time_0_to_60mph_at_gvwr_s"] = (
                    report_scenario.max_time_0_to_60mph_at_gvwr_s
                )
                if (
                    outdict["zero_to_60_loaded"] is not None
                ):  # cannot calculate if it is none (but for some reason, range and grade are handled when none)
                    report_i["delta_max_time_0_to_60mph_at_gvwr_s"] = (
                        outdict["zero_to_60_loaded"]
                        - report_scenario.max_time_0_to_60mph_at_gvwr_s
                    )

                report_i["max_time_0_to_30mph_at_gvwr_ach_s"] = outdict[
                    "zero_to_30_loaded"
                ]
                report_i["target_max_time_0_to_30mph_at_gvwr_s"] = (
                    report_scenario.max_time_0_to_30mph_at_gvwr_s
                )
                if (
                    outdict["zero_to_30_loaded"] is not None
                ):  # cannot calculate if it is none (but for some reason, range and grade are handled when none)
                    report_i["delta_max_time_0_to_30mph_at_gvwr_s"] = (
                        outdict["zero_to_30_loaded"]
                        - report_scenario.max_time_0_to_30mph_at_gvwr_s
                    )

                report_i.update(mpgge)
                # report_i["payload_capacity_loss_kg"] = outdict["payload_capacity_loss_kg"] This might be a good var to have
                report_i["payload_cap_cost_multiplier"] = veh_opp_cost_set[
                    "payload_cap_cost_multiplier"
                ]
                report_i["total_fueling_dwell_time_hr"] = sum(
                    veh_opp_cost_set["net_fueling_dwell_time_hr_per_yr"]
                )
                report_i["total_mr_downtime_hr"] = sum(
                    veh_opp_cost_set["net_mr_downtime_hr_per_yr"]
                )
                report_i["total_downtime_hr"] = sum(
                    veh_opp_cost_set["total_downtime_hr_per_yr"]
                )
                report_i["fueling_dwell_labor_cost_dol"] = disc_cost_agg.loc[
                    "fueling labor cost", "Discounted Cost [$]"
                ]
                report_i["fueling_downtime_oppy_cost_dol"] = disc_cost_agg.loc[
                    "fueling downtime cost", "Discounted Cost [$]"
                ]
                report_i["mr_downtime_oppy_cost_dol"] = disc_cost_agg.loc[
                    "MR downtime cost", "Discounted Cost [$]"
                ]
                report_i["discounted_downtime_oppy_cost_dol"] = oppy_cost_set[
                    "discounted_downtime_oppy_cost_dol"
                ]

                report_i["payload_capacity_cost_dol"] = oppy_cost_set[
                    "payload_capacity_cost_dol"
                ]
                report_i["glider_cost_dol"] = veh_cost_set["Glider"]
                report_i["fuel_converter_cost_dol"] = veh_cost_set["Fuel converter"]
                report_i["fuel_storage_cost_dol"] = veh_cost_set["Fuel Storage"]
                report_i["motor_control_power_elecs_cost_dol"] = veh_cost_set[
                    "Motor & power electronics"
                ]
                report_i["plug_cost_dol"] = veh_cost_set["Plug"]
                report_i["battery_cost_dol"] = veh_cost_set["Battery"]
                report_i["purchase_tax_dol"] = veh_cost_set["Purchase tax"]
                report_i["msrp_total_dol"] = veh_cost_set["msrp"]
                report_i["insurance_cost_dol"] = disc_cost_agg.loc[
                    "insurance", "Discounted Cost [$]"
                ]
                report_i["residual_cost_dol"] = disc_cost_agg.loc[
                    "residual cost", "Discounted Cost [$]"
                ]
                report_i["total_fuel_cost_dol"] = disc_cost_agg.loc[
                    "Fuel", "Discounted Cost [$]"
                ]

                report_i["total_maintenance_cost_dol"] = disc_cost_agg.loc[
                    "maintenance", "Discounted Cost [$]"
                ]
                report_i["discounted_tco_dol"] = disc_cost

                if outdict["design_cycle_sim_drive_record"] is not None:
                    report_i["design_cycle_EA_err"] = {
                        sdr.cyc.name: sdr.energy_audit_error
                        for sdr in outdict["design_cycle_sim_drive_record"]
                    }
                    report_i["design_cyc_trace_miss_dist_frac"] = {
                        sdr.cyc.name: sdr.trace_miss_dist_frac
                        for sdr in outdict["design_cycle_sim_drive_record"]
                    }
                    report_i["design_cyc_trace_miss_time_frac"] = {
                        sdr.cyc.name: sdr.trace_miss_time_frac
                        for sdr in outdict["design_cycle_sim_drive_record"]
                    }
                    report_i["design_cyc_trace_miss_speed_mps"] = {
                        sdr.cyc.name: sdr.trace_miss_speed_mps
                        for sdr in outdict["design_cycle_sim_drive_record"]
                    }
                if outdict["accel_sim_drive_record"] is not None:
                    report_i["accel_EA_err"] = outdict[
                        "accel_sim_drive_record"
                    ].energy_audit_error
                if outdict["accel_loaded_sim_drive_record"] is not None:
                    report_i["accel_loaded_EA_err"] = outdict[
                        "accel_loaded_sim_drive_record"
                    ].energy_audit_error
                if outdict["grade_6_sim_drive_record"] is not None:
                    report_i["grade_6_EA_err"] = outdict[
                        "grade_6_sim_drive_record"
                    ].energy_audit_error
                if outdict["grade_1p25_sim_drive_record"] is not None:
                    report_i["grade_1p25_EA_err"] = outdict[
                        "grade_1p25_sim_drive_record"
                    ].energy_audit_error

            # for all vehicles, save their final TCO TSV files
            if write_tsv:
                save_tco_files(outdict["tco_files"], resdir, scenario_name, sel, ts)

            report_i = {k: str(v) for k, v in report_i.items()}
            reports.append(report_i)
            reports_df = pd.concat([pd.DataFrame(ri, index=[0]) for ri in reports])
            try:
                # try to write output each iteration so users can view prelim results
                reports_df.to_csv(resdir / RES_FILE, index=False)
            except PermissionError:
                reports_df.to_csv(resdir / ("alternate_" + RES_FILE), index=False)
                print(f"Could not write file {resdir / RES_FILE}, file open")

        logging.info(
            f"done with selection {sel} time[s]: {opt_time} using algo: {algo} | n gens/max gens::{n_gens_used}/{n_max_gen} | {scenario_name}"
        )

        print("writing to ", resdir / RES_FILE)

    def skip_scenario(sel: int, scenario_name: str, verbose: bool = False) -> bool:
        """
        This function checks if given selection is present in exclude or look_for selections

        Args:
            sel (int): _description_
            scenario_name (str): scenario name
            verbose (bool, optional): if selected, prints out scenarios that are skipped. Defaults to False.

        Returns:
            bool: if not present, returns True; Else False
        """
        if any(ex in scenario_name for ex in exclude):
            if verbose:
                print(
                    f"skipping {sel} {exclude} has parts in scenario_name {scenario_name}"
                )
            return True
        if not any(lf in scenario_name for lf in look_for):
            if verbose:
                print(f"skipping {sel}, want to run from {look_for}")
            return True
        if sel not in selections:
            if verbose:
                print(f"skipping {sel} not in desired selections")
            return True
        return False

    if do_input_validation:
        st = time.time()
        print("sweep:: Running input validation...")
        badinputs = False
        noinputs = True
        for sel, scenario_name, optpt in zip(
            vdf.index, vdf["scenario_name"], vdf["veh_pt_type"]
        ):
            if skip_scenario(sel, scenario_name, verbose=False):
                continue
            for algo in algorithms:
                try:
                    algopart = f"algorithm: {algo.rjust(15)} or None if skipping"
                    print(
                        f"sweep:: validating input {sel}:{scenario_name}".ljust(90),
                        algopart,
                    )
                    # print(f'config: {config}')
                    input_validation(sel, optpt, algo, config)
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    badinputs = True
                    logging.exception(
                        f"sweep:: INPUT ERROR selection {sel}, {scenario_name} :: {e}",
                        stack_info=False,
                        exc_info=True,
                    )
                noinputs = False
        print(f"sweep:: Finished input validation, time [s] {round(time.time()-st)}")
        if badinputs:
            raise Exception(
                f"sweep:: input_validation failure, see log file!\n{loggingfname}"
            )
        if noinputs:
            raise Exception(
                f"sweep:: no inputs available, see log file!\n{loggingfname}"
            )

    for sel, scenario_name, optpt in zip(
        vdf.index, vdf["scenario_name"], vdf["veh_pt_type"]
    ):
        if skip_scenario(sel, scenario_name):
            continue
        scen_df = dict(sdf.loc[sel, :])
        skip_opt = scen_df.get("skip_opt", False)
        try:
            if skip_opt is True or skip_opt == "TRUE" or skip_all_opt:
                optimize(
                    sel,
                    scenario_name,
                    optpt,
                    algo="None",
                    skip_opt=True,
                    write_tsv=write_tsv,
                )
            else:
                for algo in algorithms:
                    optimize(
                        sel,
                        scenario_name,
                        optpt,
                        algo,
                        skip_opt=False,
                        write_tsv=write_tsv,
                    )
        except:
            logging.exception("Fatal Error")
            raise

    logging.info("T3CO finished")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--selections",
        help="""str: int, array or range, selections desired to run. Selections can be an int, or list of ints, or range expression. Ex: -selections 234 or -selections "[234,236,238]" or -selections "range(234, 150, 2)" """,
    )
    parser.add_argument(
        "--look-for",
        default="",
        help="str or list of strings: a string for string matching, example -look_for 'FCEV' or -look_for '[\"FCEV\", \"HEV\"]' ",
    )
    parser.add_argument(
        "--skip-all-opt",
        "--skopt",
        action="store_true",
        help="flag, if -skip_all_opt used, all runs skip optimization",
    )
    parser.add_argument(
        "--skip-input-validation",
        "--skiv",
        action="store_false",
        help="flag, if -skip_input_validation used, no pre-validation of inputs is run before sweep commences",
    )
    parser.add_argument(
        "--exclude",
        default=">{-<>-}<",
        help="str or list of strings: Overrides -look_for. a string for string matching to exclude runs, example -exclude 'FCEV' or -look_for '[\"FCEV\", \"HEV\"]'  ",
    )
    parser.add_argument(
        "--algorithms",
        "--algos",
        "--algo",
        default="NSGA2",
        help=f'str or list: Enter list of algorithms, or "ensemble" to use all, to use for optimization: {moo.ALGORITHMS} ex: -algos PatternSearch | -algos \'["PatternSearch", "NSGA2"]\' | -algos "ensemble" ',
    )
    parser.add_argument(
        "--dst-dir",
        help="directory to store results - otherwise defaults to ../tco_results/TDA_results/",
        default=None,
    )
    parser.add_argument(
        "--dir-mark",
        default="",
        help="str: name for your results directory in addition to timestamp",
    )
    parser.add_argument(
        "--file-mark", default="", help="str: name to add to your results files"
    )
    parser.add_argument(
        "--skip-save-veh",
        action="store_true",
        help="toggle result vehicle model YAML file saving off",
    )
    parser.add_argument(
        "--x-tol", default=0.001, help="parameter space tolerance"
    )  # bigger = more lax # TODO: we need to investigate if this is too big of a default
    parser.add_argument(
        "--f-tol", default=0.001, help="objective space tolerance"
    )  # bigger = more lax # TODO: we need to investigate if this is too big of a default
    parser.add_argument(
        "--n-max-gen",
        default=1000,
        help="max number of optimizer iterations regardless of algorithm",
    )
    parser.add_argument("--pop-size", default=25, help="population of each generation")
    parser.add_argument(
        "--nth-gen",
        default=1,
        help="period of generations in which to evaluate if convergence happens",
    )
    parser.add_argument(
        "--n-last",
        default=5,
        help="number of generations to look back for establishing convergence",
    )
    parser.add_argument(
        "--range-overshoot-tol",
        default=None,
        help="range overshoot tolerance, example '0.20' allows 20%% range overshoot. Default of 'None' does not constrain overshoot.",
    )
    # time-dilation-args passed to T3COProblem instantiation for optimization usage
    parser.add_argument(
        "---missed-trace-correction",
        action="store_true",
        help="bool, ex: 'sweep.py --missed_trace_correction'; activate FASTSim time-dilation to correct missed trace; default=False",
    )
    parser.add_argument(
        "--max-time-dilation",
        default=10,
        help="int, maximum time dilation factor to 'catch up' with trace; default=10  ",
    )
    parser.add_argument(
        "--min-time-dilation",
        default=0.1,
        help="float, minimum time dilation to let trace 'catch up'; default=0.1 ",
    )
    parser.add_argument(
        "--time-dilation-tol",
        default=1e-3,
        help="float, convergence criteria for time dilation; default=1e-3",
    )
    # input files
    parser.add_argument(
        "--vehicles",
        default=gl.SWEEP_PATH.parents[0]
        / "resources/inputs/demo/Demo_FY22_vehicle_model_assumptions.csv",
        help="input file for vehicles",
    )
    parser.add_argument(
        "--scenarios",
        default=gl.SWEEP_PATH.parents[0]
        / "resources/inputs/demo/Demo_FY22_scenario_assumptions.csv",
        help="input file for scenarios",
    )
    parser.add_argument(
        "--eng-curves",
        default=gl.SWEEP_PATH.parents[0]
        / "resources/auxiliary/eng_imp_cost_curves_for_demo.csv",
        help="input file for engine efficiency curves",
    )
    parser.add_argument(
        "--lw-curves",
        default=gl.SWEEP_PATH.parents[0]
        / "resources/auxiliary/matlltwt_imp_cost_curves_for_demo.csv",
        help="input file for lightweighting curves",
    )
    parser.add_argument(
        "--aero-curves",
        default=gl.SWEEP_PATH.parents[0]
        / "resources/auxiliary/aero_imp_cost_curves_for_demo.csv",
        help="input file for aerodynamics improvement curves",
    )
    parser.add_argument("--delete-me", default=True)
    parser.add_argument("--write-tsv", default=False)
    parser.add_argument(
        "--config",
        default=gl.SWEEP_PATH.parents[0] / "resources/T3COConfig.csv",
        help="input config file",
    )
    parser.add_argument(
        "--analysis-id", default="0", help="analysis selection from input config file"
    )

    print(f"gl.SWEEP_PATH: {gl.SWEEP_PATH}")

    args = parser.parse_args()
    # selections can be an int, or list of ints, or range expression
    if args.config is None:
        if args.selections is None:
            selections = -1
        elif ("[" in args.selections and "]" in args.selections) or (
            "(" in args.selections and ")" in args.selections
        ):
            selections = ast.literal_eval(args.selections)
        else:
            selections = [int(args.selections)]
        vehicles = Path(args.vehicles)
        scenarios = Path(args.scenarios)
        eng_eff_imp_curves = Path(args.eng_eff_imp_curves)
        lw_imp_curves = Path(args.lw_imp_curves)
        aero_drag_imp_curves = Path(args.aero_drag_imp_curves)
        write_tsv = args.write_tsv
        # config = None
    else:
        args.analysis_id = ast.literal_eval(args.analysis_id)
        try:
            config = rs.Config()
            config.from_file(filename=args.config, analysis_id=args.analysis_id)
        except ValueError:
            print(f"Config analysis_id not valid: {args.analysis_id}")
            config = rs.Config()
            config.validate_analysis_id(filename=args.config)
        selections = config.selections
        vehicles = gl.SWEEP_PATH.parents[0] / config.vehicle_file
        scenarios = gl.SWEEP_PATH.parents[0] / config.scenario_file
        eng_eff_imp_curves = gl.SWEEP_PATH.parents[0] / config.eng_eff_imp_curves
        lw_imp_curves = gl.SWEEP_PATH.parents[0] / config.lw_imp_curves
        aero_drag_imp_curves = gl.SWEEP_PATH.parents[0] / config.aero_drag_imp_curves
        write_tsv = config.write_tsv

    look_for = args.look_for
    exclude = args.exclude
    if "[" in look_for:
        look_for = ast.literal_eval(look_for)
    else:
        look_for = [look_for]
    if "[" in exclude:
        exclude = ast.literal_eval(exclude)
    else:
        exclude = [exclude]

    if args.algorithms == "ensemble":
        algorithms = moo.ALGORITHMS
    elif "[" in args.algorithms and "]" in args.algorithms:
        algorithms = ast.literal_eval(args.algorithms)
    else:
        algorithms = [args.algorithms]

    kwargs = {
        "selections": selections,
        "look_for": look_for,
        "exclude": exclude,
        "algo": algorithms,
        "dir_mark": args.dir_mark,
        "file_mark": args.file_mark.replace(".csv", ""),
        "skip_save_veh": args.skip_save_veh,
        "x_tol": float(args.x_tol),
        "f_tol": float(args.f_tol),
        "n_max_gen": int(args.n_max_gen),
        "pop_size": int(args.pop_size),
        "nth_gen": int(args.nth_gen),
        "n_last": int(args.n_last),
        "skip_all_opt": args.skip_all_opt
        if args.config is None
        else config.skip_all_opt,
        "do_input_validation": args.skip_input_validation,
        "range_overshoot_tol": float(args.range_overshoot_tol)
        if args.range_overshoot_tol is not None
        else None,
        "write_tsv": write_tsv,
    }
    if args.missed_trace_correction:
        kwargs.update(
            {
                "max_time_dilation": float(args.max_time_dilation),
                "min_time_dilation": float(args.min_time_dilation),
                "time_dilation_tol": float(args.time_dilation_tol),
                "missed_trace_correction": bool(args.missed_trace_correction),
            }
        )

    run_vehicle_scenarios(
        vehicles,
        scenarios,
        eng_eff_imp_curves,
        lw_imp_curves,
        aero_drag_imp_curves,
        config=config,
        **kwargs,
    )

# %%



from __future__ import annotations

from typing import Tuple
import fastsim
import pandas as pd
from t3co.objectives import fueleconomy
from t3co.run import Global as gl, run_scenario
from t3co.tco import tco_stock_emissions
from t3co.tco import tcocalc as tcocalc

# for debugging
# import importlib
# fueleconomy = importlib.reload(fueleconomy)

STD_VAR_NAMES = "stdVarNames"


def get_operating_costs(ownershipCosts: pd.DataFrame, TCO_switch: str = "DIRECT") -> pd.DataFrame:
    """
    This function creates a dataframe of operating cost from ownershipCosts dataframe based on TCO_switch ('DIRECT' or 'EFFICIENCY')

    Args:
        ownershipCosts (pd.DataFrame): Dataframe containing year-wise ownership costs estimations like Fuel, maintenance, insurance, etc
        TCO_switch (str, optional): Switch between different TCO calculations - 'DIRECT' or 'EFFICIENCY'. Defaults to 'DIRECT'.

    Returns:
        operatingCosts_df (pd.DataFrame): Dataframe containing operating cost categories based on TCO_switch
    """
    if TCO_switch == "DIRECT":
        operating_costs_categories = [
            "Fuel",
            "maintenance",
            "insurance",
            "fueling labor cost",
            "fueling downtime cost",
            "MR downtime cost",
        ]
        operatingCosts_df = ownershipCosts[
            ownershipCosts["Category"].isin(operating_costs_categories)
        ]

    elif TCO_switch == "EFFICIENCY":
        operating_costs_categories = [
            "Fuel",
            "maintenance",
            "insurance",
            "fueling labor cost",
        ]

        operatingCosts_df = ownershipCosts[
            ownershipCosts["Category"].isin(operating_costs_categories)
        ]
    return operatingCosts_df


def discounted_costs(
    scenario: run_scenario.Scenario, ownershipCosts: pd.DataFrame
) -> pd.DataFrame:
    """
    This function calculates the yearly discounted costs for each category of ownershipCosts based on scenario.discRate

    Args:
        scenario (run_scenario.Scenario): Scenario object for current selection
        ownershipCosts (pd.DataFrame): Dataframe containing year-wise ownership costs estimations like Fuel, maintenance, insurance, etc

    Returns:
        ownershipCosts (pd.DataFrame): ownershipCosts dataframe with additional 'Discounted Cost [$]' column
    """
    ownershipCosts["Discounted Cost [$]"] = ownershipCosts["Cost [$]"] / (
        1.0 + scenario.discount_rate_pct_per_yr
    ) ** (ownershipCosts["Year"] - ownershipCosts["Model Year"])
    # veh_opp_cost_set['Discounted Cost [$]'] = veh_opp_cost_set['Cost [$]'] / (1. + scenario.discount_rate_pct_per_yr) ** (veh_opp_cost_set['Year'] - veh_opp_cost_set['Model Year'])
    # veh_residual_df['Discounted Cost [$]'] = veh_residual_df['Cost [$]'] / (1. + scenario.discount_rate_pct_per_yr) ** (veh_residual_df['Year'] - veh_residual_df['Model Year'])

    return ownershipCosts


def calc_discountedTCO(
    scenario: run_scenario.Scenario,
    discounted_costs_df: pd.DataFrame,
    veh_cost_set: dict,
    veh_opp_cost_set: dict,
    sim_drive: fastsim.simdrive.SimDrive,
    TCO_switch: str = "DIRECT",
) -> Tuple[float, dict, dict]:
    """
    This function calculates the discounted Total Cost of Ownerhip (discounted to account for time-value of money).
    There are two methods to calculate discounted TCO - 'DIRECT' and 'EFFICIENCY'

    Args:
        scenario (run_scenario.Scenario): Scenario object for current selection
        discounted_costs_df (pd.DataFrame): discounted operating costs dataframe
        veh_cost_set (dict): Dictionary containing MSRP breakdown
        veh_opp_cost_set (dict): Dictionary containing opportunity costs breakdown
        sim_drive (fastsim.simdrive.SimDrive): FASTSim.simdrive.SimDrive object containing inputs and outputs from vehicle simulation over a cycle
        TCO_switch (str, optional): Switch between different TCO calculations - 'DIRECT' or 'EFFICIENCY'. Defaults to 'DIRECT'.

    Returns:
        discounted_tco_dol (float): Discounted Total Cost of Ownership value
        oppy_cost_dol_set (dict): Dictionary containing discounted opportunity costs breakdown
        veh_oper_cost_set (dict): Dictionary containing discounted operating costs breakdown
    """
    operatingCosts_df = get_operating_costs(discounted_costs_df, TCO_switch)
    totaloperatingCosts_df = operatingCosts_df.groupby(["Year", "Model Year"]).sum(
        "Cost [$]"
    )

    payloadmultiplier = veh_opp_cost_set["payload_cap_cost_multiplier"] or 1
    disc_operating_costs = totaloperatingCosts_df["Discounted Cost [$]"].sum()
    disc_residual_costs = discounted_costs_df[
        discounted_costs_df["Category"].isin(["residual cost"])
    ]["Discounted Cost [$]"].sum()
    disc_opportunity_costs = discounted_costs_df[
        discounted_costs_df["Category"].isin(
            ["fueling downtime cost", "MR downtime cost"]
        )
    ]["Discounted Cost [$]"].sum()
    sum_operating_costsdf = operatingCosts_df.groupby("Category", as_index=False).sum(
        "Discounted Cost [$]"
    )
    veh_oper_cost_set = dict(
        zip(
            sum_operating_costsdf["Category"],
            sum_operating_costsdf["Discounted Cost [$]"],
        )
    )
    if TCO_switch == "DIRECT":
        discounted_tco_dol = payloadmultiplier * (
            veh_cost_set["msrp"]
            + veh_cost_set["Purchase tax"]
            + disc_operating_costs
            + disc_residual_costs
        )
        payload_capacity_cost = (
            (payloadmultiplier - 1) / payloadmultiplier * discounted_tco_dol
        )
        oppy_cost_dol_set = {
            "discounted_downtime_oppy_cost_dol": disc_opportunity_costs,
            "payload_capacity_cost_dol": payload_capacity_cost,
        }

    elif TCO_switch == "EFFICIENCY":
        disc_VMT_sum = sum(
            [
                scenario.vmt[i] / (1 + scenario.discount_rate_pct_per_yr) ** (i)
                for i in range(scenario.vehicle_life_yr)
            ]
        )
        disc_downtime_sum = sum(
            [
                veh_opp_cost_set["total_downtime_hrPerYr"][i]
                / (1 + scenario.discount_rate_pct_per_yr) ** (i + 1)
                for i in range(scenario.vehicle_life_yr)
            ]
        )
        avg_speed_mph = (
            sum(sim_drive.cyc.mps) / max(sim_drive.cyc.time_s) * gl.mps_to_mph
        )
        downtime_efficiency = 1 / (1 + avg_speed_mph * disc_downtime_sum / disc_VMT_sum)
        # print(f'downtime_efficiency = {downtime_efficiency}')
        discounted_tco_dol = payloadmultiplier * (
            (veh_cost_set["msrp"] + veh_cost_set["Purchase tax"] + disc_operating_costs)
            / downtime_efficiency
            + disc_residual_costs
        )
        discounted_downtime_oppy_cost_dol = (
            veh_cost_set["msrp"]
            + veh_cost_set["Purchase tax"]
            + disc_operating_costs
            + disc_opportunity_costs
        ) * (1 / downtime_efficiency - 1)
        payload_capacity_cost = (
            (payloadmultiplier - 1) / payloadmultiplier * discounted_tco_dol
        )
        oppy_cost_dol_set = {
            "discounted_downtime_oppy_cost_dol": discounted_downtime_oppy_cost_dol,
            "payload_capacity_cost_dol": payload_capacity_cost,
        }

    return discounted_tco_dol, oppy_cost_dol_set, veh_oper_cost_set


def get_tco_of_vehicle(
    vehicle: fastsim.vehicle.Vehicle,
    range_cyc: fastsim.cycle.Cycle,
    scenario: run_scenario.Scenario,
    write_tsv: bool = False,
) -> Tuple[
    float,
    float,
    dict,
    pd.DataFrame,
    pd.DataFrame,
    dict,
    dict,
    fastsim.simdrive.SimDrive,
    dict,
    dict,
    dict,
]:
    """
    This function calculates the Total Cost of Ownership of a vehicle and scenario for a given cycle. The three main components are:
    - Opportunity Costs
    - MSRP
    - Operating Costs

    Args:
        vehicle (fastsim.vehicle.Vehicle): FASTSim vehicle object of selected vehicle
        range_cyc (fastsim.cycle.Cycle): FASTSim range cycle object
        scenario (run_scenario.Scenario): Scenario object for current selection
        write_tsv (bool, optional): if True, save intermediate files as TSV. Defaults to False.

    Returns:
        tot_cost_dol (float): TCO in dollars
        discounted_tco_dol (float): discounted TCO in dollars
        oppy_cost_set (dict): Dictionary of opportunity cost breakdown
        ownership_costs_df (pd.DataFrame): Ownerhip Costs dataframe containing different categories per year
        discounted_costs_df (pd.DataFrame): discounted Ownerhip Costs dataframe containing different categories per year
        mpgge (dict): Dictionary containing MPGGEs
        veh_cost_set (dict): Dictionary containing MSRP breakdown
        design_cycle_sdr (fastsim.simdrive.SimDrive): FASTSim SimDrive object for design drivecycle
        veh_oper_cost_set (dict): Dictionary containing operating costs breakdown
        veh_opp_cost_set (dict): Dictionary containing opportunity costs breakdown
        tco_files (dict): Dictionary containing TCO intermediate dataframes
    """

    mpgge, sim_drives = fueleconomy.get_mpgge(range_cyc, vehicle, scenario)
    range_dict = fueleconomy.get_range_mi(mpgge, vehicle, scenario)
    veh_opp_cost_set = tcocalc.calculate_opp_costs(vehicle, scenario, range_dict)
    veh_cost_set = tcocalc.calculate_dollar_cost(vehicle, scenario)
    veh_eff_df = tcocalc.fill_fuel_eff_file(vehicle, scenario, mpgge)
    veh_exp_df = tcocalc.fill_veh_expense_file(scenario, veh_cost_set)
    veh_spt_df = tcocalc.fill_fuel_split_tsv(vehicle, scenario, mpgge)
    veh_txp_df = tcocalc.fill_trav_exp_tsv(vehicle, scenario)
    veh_fxp_df = tcocalc.fill_fuel_expense_tsv(vehicle, scenario)
    veh_shr_df = tcocalc.fill_market_share_tsv(scenario)
    ann_trv_df = tcocalc.fill_annual_tsv(scenario)
    reg_sls_df = tcocalc.fill_reg_sales_tsv(scenario)
    survivl_df = tcocalc.fill_survival_tsv(scenario)
    veh_insurance_df = tcocalc.fill_insurance_tsv(scenario, veh_cost_set)
    veh_residual_df = tcocalc.fill_residual_cost_tsc(vehicle, scenario, veh_cost_set)
    veh_downtime_labor_df = tcocalc.fill_downtimelabor_cost_tsv(
        scenario, veh_opp_cost_set
    )

    # emission_df = pd.read_csv(gl.TCO_INTERMEDIATES / gl.EMISSION_RATE_TSV, index_col=None, header=0, sep='\t')
    emission_df = None

    # run stock model
    stock, emissions, ownership_costs_df = tco_stock_emissions.stockModel(
        reg_sls_df,
        veh_shr_df,
        survivl_df,
        ann_trv_df,
        veh_spt_df,
        veh_eff_df,
        emission_df,
        veh_exp_df,
        veh_txp_df,
        veh_fxp_df,
        veh_insurance_df,
        veh_residual_df,
        veh_downtime_labor_df,
        write_files=gl.write_files,
    )

    # discountRate = float(scenario.discount_rate_pct_per_yr)
    # discounted_costs_df = DCF(ownership_costs_df.copy(), rate=discountRate)
    # print(discounted_costs_df)
    discounted_costs_df = discounted_costs(scenario, ownership_costs_df)
    # should only be one vocation in these files but this as good a thing to aggregate on as any
    tot_cost_dol = discounted_costs_df["Cost [$]"].sum()

    # discounted_tco_dol, discounted_downtime_oppy_cost_dol, veh_oper_cost_set = calc_discountedTCO(scenario, discounted_costs_df, veh_cost_set, veh_opp_cost_set, sim_drives[-1], TCO_switch = 'DIRECT')
    # print(f'New disc DIRECT TCO: {discounted_tco_dol}')
    discounted_tco_dol, oppy_cost_set, veh_oper_cost_set = calc_discountedTCO(
        scenario,
        discounted_costs_df,
        veh_cost_set,
        veh_opp_cost_set,
        sim_drives[-1],
        TCO_switch="DIRECT",
    )
    # print(f'New disc EFFICIENCY TCO: {discounted_tco_dol}')

    # if veh_opp_cost_set['payload_cap_cost_multiplier'] is not None:
    #     discounted_costs_df["Payload Corrected Discounted Cost [$]"] = \
    #       discounted_costs_df["Discounted Cost [$]"] * veh_opp_cost_set['payload_cap_cost_multiplier']
    #     disc_cost = discounted_costs_df["Payload Corrected Discounted Cost [$]"].sum()

    # write output files
    tco_files = {}
    if write_tsv:
        tco_files = {
            "veh_eff_df": veh_eff_df,
            "veh_exp_df": veh_exp_df,
            "veh_txp_df": veh_txp_df,
            "veh_shr_df": veh_shr_df,
            "veh_fxp_df": veh_fxp_df,
            "ann_trv_df": ann_trv_df,
            "reg_sls_df": reg_sls_df,
            "survivl_df": survivl_df,
            "veh_spt_df": veh_spt_df,
            "stock": stock,
            "ownership_costs_df": ownership_costs_df,
            "discounted_costs_df": discounted_costs_df,
        }

    return (
        tot_cost_dol,
        discounted_tco_dol,
        oppy_cost_set,
        ownership_costs_df,
        discounted_costs_df,
        mpgge,
        veh_cost_set,
        sim_drives,
        veh_oper_cost_set,
        veh_opp_cost_set,
        tco_files,
    )

    # print(operatingCosts_df)



"""
Module for testing moo.  Folder structure, file name, and code 
are written to be compliant with python's unittest package, but 
main() can be called via importing
"""


import unittest 
from t3co import Global as gl


class TestMoo(unittest.TestCase):
    def test_moo(self):
        self.assertTrue(True) # TODO: actually build a test here

if __name__ == '__main__':
    unittest.main()